<?php
/*
 * Micro Album v2.2
 * Gaby Vanhegan <gaby@vanhegan.net> 20080423
 * Extremely lightweight PHP based image gallery.  Very simple,
 * not fancy or complicated.  This should work in it's favour
 * for speed and security.
 *
 * Thanks to pbob for a bug fix in _image_get_generated();
 *
 * Features: 
 * - Roating thumbnails/previews
 * - Listing albums by name/date/custom order/reverse
 * - Listing images by name/date/custom order/reverse
 * - Password protected admin area
 * - Single page setup, very quick simple and easy to use
 * - Relatively tidy design
 * - Build images on the fly or all in a batch
 * - Built in a handy class format for easy extensions
 * - Tweak preview and thumbnail build for speed/quality
 * - Option to output data in JSON format
 */

/**
 * ---=== The following options must be set correctly ===---
 */
$config	= array();	// Leave me here please!

/**
 * The folder where this script is located, and the folders that
 * make up the individual albums are found.
 */
$config[ "base_folder" ]			= "/Users/gaby/Sites/microalbum/";

/**
 * The folder where generated thumbnails and previews will go.  This
 * !!! THIS MUST BE INSIDE THE BASE FOLDER !!!
 */
$config[ "gen_folder" ]				= "generated";

/**
 * The password that gives access to the edit pages
 */
$config[ "admin_pass" ]				= "changeme!";

/**
 * ---=== Everything below this point is optional ===---
 */

// The name of this album system
$config[ "page_title" ]				= "Photo Album";

// Optional homepage link, in the title nav bar.  Set to false for none.
$config[ "home_link" ]				= "vanhegan.net";
$config[ "home_url" ]				= "http://vanhegan.net/";

// Custom CSS, HTML header and HTML footer.  Set to false for none.
// Header and footer files muse be within the base folder.
$config[ "custom_css" ]				= false;
$config[ "custom_header" ]			= false;
$config[ "custom_footer" ]			= false;

// How many thumbnails to display per page
$config[ "thumb_per_row" ]			= 4;
$config[ "thumb_per_page" ]		= 12;

// How many seconds does the slideshow display each image for?
$config[ "slide_time" ]				= 4;

// How many seconds does the slideshow display each image for?
$config[ "most_popular" ]			= 100;

// Do we build thumbnails and previews on the fly?
// WARNING: If these are set to true then the first time the album
// is loaded or viewed will be slower as the images get built for
// the first time.
$config[ "live_build" ]				= true;

/*
 * ---=== Advanced only settings below this line ===---
 */

// Generated image dimensions.  Changing these will mean that all the
// generated images may have to be rebuilt.
$config[ "thumb_x" ]					= 160;
$config[ "thumb_y" ]					= 120;
$config[ "preview_x" ]				= 640;
$config[ "preview_y" ]				= 480;

// Maximum length of captions in thumbnail mode
$config[ "thumb_caption_len"]		= 26;

// Allowed file types, ordering methods
$config[ "valid_types" ]		= array( "jpg", "jpeg", "gif", "png", "bmp" );
$config[ "valid_orders" ]		= array( "date", "file", "custom" );

set_time_limit( 3600 );

// Do the doings please
$album			= new MicroAlbum( $config, $_GET, $_POST );
$album->render();

class MicroAlbum {
	
	/**
	 * @var	array 	User-specified program parameters
	 */
	var $params;
	
	/**
	 * @var	string 	The built output
	 */
	var $_output;
	
	/**
	 * @var	array 	List of links in the title
	 */
	var $_html_title 	= array();
	
	/**
	 * @var	array 	List of items in the navigation bar
	 */
	var $_html_nav		= array();
	
	/**
	 * @var	array 	List of items in the admin section
	 */
	var $_html_admin	= array();
	
	/**
	 * @var	array 	Error messages that need displaying
	 */
	var $_msg_error	= array();
	
	/**
	 * @var	array 	Success messages to be displayed
	 */
	var $_msg_good		= array();
	
	/**
	 * @var	array 	Loaded data about the currently viewing album
	 */
	var $_album_data;
	
	/**
	 * @var	string 	Name of the current album we're in
	 */
	var $_album_name;
	
	/**
	 * @var	boolean	Are we working in an album?
	 */
	var $_in_album		= false;
	
	/**
	 * @var	boolean	Are we in slideshow mode?
	 */
	var $_in_show		= false;
	
	/**
	 * @var	array 	Summary data about all the albums
	 */	
	var $_summary_data;
	
	/**
	 * @var	string 	The folder that contains all the albums
	 */
	var $_folder_base;
	
	/**
	 * @var	string 	The sub-folder that contains generated thumbs/previews
	 */
	var $_folder_gen;
	
	/**
	 * @var	array 	List of log messages
	 */
	var $_debug_log	= array();
	
	/**
	 * @var	string 	microtime( true ) when we started building
	 */
	var $_page_time;
	
	/**
	 * @var	array 	Contents of the $_GET server variable
	 */
	var $_get;

	/**
	 * @var	array 	Contents of the $_POST server variable
	 */
	var $_post;
	
	/**
	 * Create a new MicroAlbum
	 * @param	array 	$params	User supplied config parameters
	 * @param	array 	$get		Array of server variables from this HTTP GET
	 * @param	array 	$post		Array of server variables from this HTTP POST
	 * @access	public
	 */
	public function MicroAlbum ( $params, $get, $post ) {
		
		// Mark the time we were called into existance
		list( $usec, $sec )	= explode( " ", microtime() );
		$this->_page_time		= ( (float)$usec + (float)$sec );
		
		// Sanitize the GET input
		foreach ( $get  as $k => $v ) { 
			$this->_get[ $k ]  = stripslashes( $v );
		}
		
		// Sanitize the POST input, mindful that it may have sub-arrays
		foreach ( $post as $k => $v ) { 
			if ( is_array( $v ) ) {
				$new_list	= array();
				foreach ( $v as $val ) {
					array_push( $new_list, stripslashes( $val ) );
				}
				$this->_post[ $k ]	= $new_list;
			}
			else {
				$this->_post[ $k ] = stripslashes( $v ); 
			}
		}
		
		$this->params	= array();
		
		$this->params[ "page_title" ]			= "Photo Album";
		$this->params[ "home_link" ]			= "vanhegan.net";
		$this->params[ "home_url" ]			= "http://vanhegan.net/";
		
		$this->params[ "custom_css" ]			= false;
		$this->params[ "custom_header" ]		= false;
		$this->params[ "custom_footer" ]		= false;
		
		$this->params[ "thumb_per_row" ]		= 4;
		$this->params[ "thumb_per_page" ]	= 12;
		
		$this->params[ "slide_time" ]			= 4;
		
		$this->params[ "most_popular" ]		= 50;
		
		$this->params[ "live_build" ]			= true;
		
		// What joins the title links?
		$this->params[ "title_join" ]			= " &gt; ";
		
		// What joins the navigation links?
		$this->params[ "nav_join" ]			= " ";
		$this->params[ "nav_spacer" ]			= "::";

		$this->params[ "thumb_x" ]				= 160;
		$this->params[ "thumb_y" ]				= 120;
		$this->params[ "thumb_quality" ]		= 2.0;
		
		$this->params[ "preview_x" ]			= 640;
		$this->params[ "preview_y" ]			= 480;
		$this->params[ "preview_quality" ]	= 2.5;
		
		$this->params[ "thumb_caption_len"]	= 26;
		$this->params[ "valid_types" ]		= array( 
															"jpg", 
															"jpeg", 
															"gif", 
															"png", 
															"bmp" 
														);
		$this->params[ "valid_orders" ]		= array( 
															"date", 
															"file", 
															"custom"
														);

		$this->params["debug"]					= 0;
		$this->params["config_check"]			= false;

		// Override the defaults with the given parameters
		foreach ( $params as $key => $val ) { $this->params[ $key ] = $val; }
		
		// File version, data format version
		$this->params["version"]				= "2.4";		// Program version
		$this->params["data_version"]			= "1.4b";	// Data format version
		
		// Link to ourselves
		$this->params["self"]					= $_SERVER["PHP_SELF"];
		
		// Are we rebuilding any albums?
		$this->params["album_rebuild"]	= array();
		
		// Width of individual table cells in %
		$percent		= ( 100 / $this->params["thumb_per_row"] );
		$this->params["td_width"]	= sprintf( "%u", $percent );
		
		// Base folder that albums live in
		$this->_folder_base 	= realpath( $this->params["base_folder"] );
		
		// Generated image folder
		$this->_folder_gen 	= $this->params[ "gen_folder" ];
		
		// Base64 encoded tick.jpeg and cross.jpeg
		$tick		= "/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAgGBgcGBQgHBwcJCQgKDBQND"
					. "AsLDBkSEw8UHRofHh0aHBwgJC4nICIsIxwcKDcpLDAxNDQ0Hyc5PTgyPC"
					. "4zNDL/2wBDAQkJCQwLDBgNDRgyIRwhMjIyMjIyMjIyMjIyMjIyMjIyMjI"			
					. "yMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjL/wAARCAAUABQDASIA"
					. "AhEBAxEB/8QAGAABAQADAAAAAAAAAAAAAAAAAAUEBgf/xAAmEAAABQQBA"
					. "gcAAAAAAAAAAAAAAQIDBAUGETEhE0EiMlFhodHh/8QAFQEBAQAAAAAAAA"
					. "AAAAAAAAAAAAL/xAAXEQEBAQEAAAAAAAAAAAAAAAAAAREC/9oADAMBAAI"
					. "RAxEAPwDv4nVGuU2kyoUabLQy9Nc6TCFbWr60WfchjXO3XjpZO268ymay"
					. "snDZdSRlISW28n5c+vyWxGtW1JRT13LcykyK6+XgRtENHZCO2eeT/TObb"
					. "uRF6u5I3MAAUsAAAAAAH//Z";
		
		// Used when rebuilding the images
		$cross	= "/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAgGBgcGBQgHBwcJCQgKDBQND"
					. "AsLDBkSEw8UHRofHh0aHBwgJC4nICIsIxwcKDcpLDAxNDQ0Hyc5PTgyPC"
					. "4zNDL/2wBDAQkJCQwLDBgNDRgyIRwhMjIyMjIyMjIyMjIyMjIyMjIyMjI"
					. "yMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjL/wAARCAAUABQDASIA"
					. "AhEBAxEB/8QAFwABAQEBAAAAAAAAAAAAAAAAAAcCA//EACYQAAEEAgICA"
					. "QQDAAAAAAAAAAECAwQRACEFBhJBMQcTFFEjcYH/xAAYAQADAQEAAAAAAA"
					. "AAAAAAAAAABAYBAv/EAB4RAAEEAgMBAAAAAAAAAAAAAAEAAgMRITEEBbE"
					. "T/9oADAMBAAIRAxEAPwCu9m7W31eRAVLhPLgSVqQ7KQbDB1Vit3v2NA1f"
					. "xmOe7nB4qNE/CCeSmzlJESMw6P5ATXl5boer/f8Atce/8vEgdfVBeiiZK"
					. "5C2I0UoKvuK1uhvVj43dVk+i8Byv01mQOwy4jc2MUeEkNi1RVK1QP7qh5"
					. "fFkp9glOWV7XEDXio+v6/jTwMkkFPzQuvpXmcGt6GVaGFOrjtLfbDTykA"
					. "rbSryCVVsA0Lo+6xmIctqfBjzGCSy+2l1skUSlQsa/o4xsaU84EOIIpd8"
					. "YxmrlMYxghf/2Q==";
		
		// No icon please
		$this->_icon_tick		= $tick;
		$this->_icon_cross	= $cross;
		
		if ( $this->params["debug"] > 0 ) {
			$this->debug( "Debug level: " . $this->params["debug"], 1 );
		}
		
		// Either of these states and we don't pre-load the album
		if ( $this->_get["rebuild"] == "thumb" ) {
			// Nothing
		}
		else if ( $this->_get["rebuild"] == "preview" ) {
			// Nothing
		}
		
		// If we have an album
		else if ( $this->_get["album"] != "" ) {
			
			// No such album?
			if ( !( $this->_album_exists( $this->_get["album"] ) ) ) { 
				$this->exit_error( "No such album" );
				return false;
			}
			
			// Flag as being in an album
			$this->_in_album		= true;
			
			// Store the album name
			$this->_album_name	= $this->_get["album"];
			
			// Store the album data
			$this->_album_data	= $this->_album_data_load( $this->_album_name );
			
			// A link back to the homepage for this album
			$this->params["album_url"]	= $this->params["self"]
												. "?album=" . urlencode( $this->_get["album"] );
			
			// Are we in a slideshow mode?
			if ( $this->_get["slideshow"] != "" ) {
				$this->_in_show 	= true;
			}
		}
		
		// Do we check the configuration?
		if ( $this->params["config_check"] != false ) { 
			$this->_config_check();
		}
	}
	
	/**
	 * Draw the album to screen in whatever mode we're in
	 * @access	public
	 */
	public function render ( ) {
		
		// Reset these list holders
		$this->_output			= "";
		$this->_html_nav		= array();
		$this->_html_admin	= array();
		$this->_html_title	= array();
		
		$format					= $this->_get["format"];
		
		// No album?
		if ( $this->_get["album"] == "" ) {

			if ( $this->_get["edit"] == "sort" ) {
				$this->_render_album_list_sort();
			}
			else if ( $this->_get["summary"] == "popular" ) {
				$this->_render_album_popular( $format );
			}
			else {
				$this->_render_album_list( $format );
			}
			
		}
		
		// Are we unlocking an album?
		else if ( $this->_get["unlock"] != "" ) {
			$this->_render_album_unlock();
		}
		
		// Editing an album?
		else if ( $this->_get["edit"] != "" ) {
			
			// What 
			if ( $this->_get["edit"] == "sort" ) {
				$this->_render_album_edit_sort();
			}
			else if ( $this->_get["edit"] == "rotate" ) {
				$this->_render_album_edit_rotate();
			}
			else {
				$this->_render_album_edit_captions();
			}
		}
		
		// Rebuilding the thumb/preview cache?
		else if ( $this->_get["rebuild"] != "" ) {

			if ( $this->_get["rebuild"] == "thumb" ) {
				$this->_render_image_rebuild();
			}
			else if ( $this->_get["rebuild"] == "preview" ) {
				$this->_render_image_rebuild();
			}
			else {
				$this->_render_album_rebuild();
			}
		}
		
		// Fetching an album as a zip?
		else if ( $this->_get["fetch"] != "" ) {
			$this->_render_album_fetch();
		}
		
		// No specific place in an album?
		else if ( $this->_get["place"] == "" ) {
			$this->_render_album_view( $format );
		}
		
		// Looking at one picture
		else {
			$this->_render_image_view( $format );
		}
		
		// JSON as a URL string format?
		if ( $this->_get["format"] == "jsonurl" ) {
			$json	= $this->json_encode( $this->_output );
			print "json=" . rawurlencode( $json ) . "\n";
		}
		
		// Just the actual raw JSON?
		else if ( $this->_get["format"] == "json" ) {
			print $this->json_encode( $this->_output );
		}
		
		// Default to HTML format
		else {
			print	  $this->_html_head() 
					. $this->_output 
					. $this->_html_foot();
		}
		
		// If we're in an album then we should save the album data when done
		if ( $this->_in_album ) {
			$this->_album_data_save( $this->_album_name, $this->_album_data );
		}
		
		return true;
	}

	/**
	 * Render the album list view
	 * @param	string	$format	Format to output in
	 * @access	private
	 */
	private function _render_album_list ( $format = "html" ) {
		
		$count			= 0;
		$data				= $this->_album_summary_load();
		$self				= $this->params["self"];
		$td_width		= $this->params["td_width"];
		$base_folder	= $this->_folder_base;
		$list				= $this->_album_list();

		// What's our sort order?
		if ( $data["display_by"] == "date" ) {
			
			// Build a pre-list
			$sorted_date	= array();
			
			foreach ( $list as $album ) {
				$item				= array();
				$item["album"]	= $album;
				$item["file"]	= $base_folder . "/" . $album;
				$item["mtime"]	= filemtime( $item["file"] );
				array_push( $sorted_date, $item );
			}
			
			// Sort by mtime
			uasort( $sorted_date, array( "MicroAlbum", "_cmp_mtime" ) );
			
			// Pick the list back out
			$list	= array();
			foreach ( $sorted_date as $item ) {
				array_push( $list, $item["album"] );
			}
		}
		
		// Sorted by filename?
		else if ( $data["display_by"] == "file" )   { 
			natcasesort( $list );
		}
		
		// Custom order?
		else if ( $data["display_by"] == "custom" ) { 
			$list = $data["custom_order"]; 
		}
	
		// Is it reversed?
		if ( $data["display_reverse"] == '1' ) { 
			$list = array_reverse( $list ); 
		}
		
		// JSON format returns the object
		if ( ( $format == "json" ) or ( $format == "jsonurl" ) ) {
			$data				= array();
			$data["list"]	= $list;
			$this->_output	= $data;
			return true;
		}
		
		// Create the list of navigation links
		$str		= count( $list ) . " albums";
		array_push( $this->_html_nav, $str );
		
		// Spacer
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		
		// Most Viewed link
		$str		= "<a href=\"$self?summary=popular\" title=\"Most viewed\">"
					. "Most viewed images"
					. "</a>";
		array_push( $this->_html_nav, $str );

		$str		= "<a href=\"$self?edit=sort\" accesskey=\"e\">Edit</a>";
		array_push( $this->_html_admin, $str );

		// Start the output
		$html		= "<table border=\"0\" width=\"100%\">\n"
					. "<tr>\n";

		// Create the page by scanning through each album and picking one
		// image at random from it
		foreach ( $list as $album ) {
			
			// Fetch a random image
			$image_data		= $this->_album_get_random( $album, "thumb" );

			// What time was the album last modified?
			$album_time		= filemtime( "$base_folder/$album" );
			
			// Build the page
			$html	.= "<td align=\"center\" width=\"$td_width%\">"
					. "<div class=\"thumb\">"
					. "<a "
					. "href=\"$self?album=" . urlencode( $album ) . "\" "
					. "title=\"View $album\" "
					. ">"
						. $image_data["thumb_html"]
						. "<br />"
						. str_replace( "_", " ", $album )
					. "</a>"
					. "<br />\n"
					. "Updated " . date( "jS F Y", $album_time )
					. "</div>\n"
					. "</td>\n"
					. "";
					
			if ( ( ++$count % $this->params["thumb_per_row"] ) == 0 ) {
				$html		.= "</tr>\n<tr>\n";
			}
		}
		$html				.= "</tr>\n</table>\n";
		
		$this->_output	= $html;		
	}
	
	/**
	 * Render the album list sorting view
	 * @access	private
	 */
	private function _render_album_list_sort ( ) {

		// Check for the password
		$this->_password_check( $this->params["admin_pass"] );
		
		// Basic variables we will need
		$self				= $this->params["self"];
		
		// Link back to the front page
		$link_back		= "<a href=\"$self\" accesskey=\"b\">&lt;&lt; Back</a>";
		array_push( $this->_html_admin, $link_back );

		// Pull all the album data out
		$album_data		= $this->_album_summary_load();
		
		// By default our list os the custom ordering
		$album_list		= $album_data["custom_order"];
		
		$order_valid	= $this->params["valid_orders"];

		// We're in action mode
		if ( $this->_post["go"] == '1' ) {
			
			// Provided there are items selected in the list
			if ( is_array( $this->_post["custom"] ) ) {
								
				// The items we're moving
				$move			= $this->_post["custom"];
				$dir			= 1;

				// Moving them up or down?
				if      ( $this->_post["up"]   ) { $dir = -1; }
				else if ( $this->_post["down"] ) { $dir = 1;  }

				// Re-order the list
				$new_list	= $this->list_arrange( $album_list, $move, $dir );
				
				// Drop it in place
				$album_data["custom_order"]	= $new_list;
				
				// Save the list
				$this->_album_summary_save( $album_data );
			}
			
			// Default to date if an invalid type is selected
			$order_by		= "date";
			if ( in_array( $this->_post["sortby"], $order_valid ) ) {
				$order_by	= $this->_post["sortby"];
			}
			
			// Save these details
			$album_data["display_by"]			= $order_by;
			$album_data["display_reverse"]	= $this->_post["reverse"];
			
			// Save the data
			$this->_album_summary_save( $album_data );
			
			// Reload the data
			$album_data	= $this->_album_summary_load();
			$album_list	= $album_data["custom_order"];
			
			$str	= "<span class=\"good\">Sort order updated</span>";
			array_unshift( $this->_html_nav, $str );
		}
		
		// We're in load mode
		else {
			$this->_post["sortby"]	= $album_data["display_by"];
			$this->_post["reverse"]	= $album_data["display_reverse"];
		}
		
		// The visible select list
		$select	= '<select name="custom[]" multiple="multiple" size="20">';
		foreach ( $album_list as $album ) {
			$key			= $album;
			$value		= $album;
			$selected	= "";
			if ( is_array( $this->_post["custom"] ) ) {
				if ( in_array( $album, $this->_post["custom"] ) ) { 
					$selected = "selected"; 
				}
			}
			$select	.= "<option value=\"$value\"$selected>$key</option>\n";
		}
		$select	.= "</select>";
		
		$html	= "<form action=\"$self?edit=sort\" method=\"POST\">\n"
				. $this->html_hidden( "go", "1" )

				. "<table class=\"normal\" border=\"0\" width=\"100%\">\n"
				. "<tr>\n"
					. "<td width=\"50%\">\n"
					. "<h3>Sort By</h3>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "date", " Date updated" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "file", " Album name" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "custom", " Custom order" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_checkbox( "reverse", "1", " Reverse order" ) 
					. "</p>\n"
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "</td>\n"
					. "<td width=\"50%\">\n"
					. "<h3>Custom Order</h3>\n"
					. "<p>"
						. $select
					. "</p>\n"
					. "<p>Move selected "
						. $this->html_submit( "up", "&uarr;" )
						. " "
						. $this->html_submit( "down", "&darr;" )
					. "</p>\n"
					. "</td>\n"
				. "</tr>\n"
				. "</table>\n"
				. "</form>\n";
		
		$this->_output	= $html;
	}
	
	/**
	 * Render the album list view
	 * @param	string	$format	Format to output in
	 * @access	private
	 */
	private function _render_album_popular ( $format = "html" ) {

		// List all the albums
		$album_list	= $this->_album_list();
		
		// The placeholder for all images
		$all_images	= array();
		
		foreach ( $album_list as $album ) {
			
			$album_data	= $this->_album_data_load( $album );
			
			$display_by	= $album_data[ "display_by" ];
			$reverse		= $album_data[ "display_reverse" ];
			
			// Get the image list sorted accordingly
			$image_list	= $album_data[ "sorted" ][ $display_by ];
			
			// Do we reverse it?
			if ( $reverse ) { $image_list = array_reverse( $image_list ); }
			
			foreach ( $image_list as $place => $image ) {
				$views	= sprintf( "%u", $album_data[ $image ][ "views" ] );
				$item					= array();
				$item["place"]		= $place;
				$item["image"]		= $image;
				$item["album"]		= $album;
				$item["views"]		= $views;
				$item["caption"]	= $the_album[ $image ][ "caption" ];
				array_push( $all_images, $item );
			}
		}

		// Sort the list of images by views
		usort( $all_images, array( "MicroAlbum", "_cmp_views" ) );
		$all_images	= array_reverse( $all_images );

		// Either take the first X or the whole list, whichever is smaller
		$end		= min( $this->params["most_popular"], count( $all_images ) );

		// Link back to the home page
		$url		= $this->params["self"];
		$str		= "<a href=\"$url\" title=\"Album listing\">"
					. "&lt;&lt; Back"
					. "</a>";
		array_push( $this->_html_nav, $str );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		$str		= "$end most viewed images";
		array_push( $this->_html_nav, $str );

		// Title link
		$str		= "Popular Images";
		array_push( $this->_html_title, $str );
		
		// Off we go
		$html		= "<table border=\"0\" width=\"100%\">\n"
					. "<tr>\n";

		// Chop out the list of images we want
		$sub_list	= array_slice( $all_images, 0, $end );
		$count		= 0;
		$td_width	= $this->params["td_width"];
		$self			= $this->params["self"];
		
		// Output these images now
		foreach ( $sub_list as $the_file ) {
			
			$the_place	= $the_file["place"];
			$the_album	= $the_file["album"];
			$the_image	= $the_file["image"];
			$the_views	= $the_file["views"];
			$the_cap		= $the_file["caption"];
			$the_name	= ( $the_cap == "" ) ? $the_image : $the_cap;				
			$image_data	= $this->_image_get( $the_album, $the_image, "thumb" );
			$the_url 	= $self
							. "?album=" . urlencode( $the_album )
							. "&place=$the_place"
							. "";
			
			$html			.= "<td align=\"center\" width=\"$td_width%\">"
							. "<div class=\"thumb\">"
							. "<a href=\"$the_url\" title=\"View $the_image\">"
							. $image_data["thumb_html"]
							. "<br />"
							. str_replace( "_", " ", $the_name )
							. "</a>"
							. "<br />\n"
							. "$the_album, $the_views views"
							. "</div>\n"
							. "</td>\n"
							. "";
			if ( ( ++$count % $this->params["thumb_per_row"] ) == 0 ) {
				$html		.= "</tr>\n<tr>\n";
			}
		}
		$html	.= "</tr>\n</table>\n";	
		
		$this->_output	= $html;
	}
	
	/**
	 * Render the album unlock view
	 * @access	private
	 */
	private function _render_album_unlock ( ) {
		
		// The lock file here
		$lock_file		= $this->_folder_base . "/"
							. $this->_folder_gen . "/"
							. $this->_album_name . "/"
							. "album.dat.lock";
		if ( file_exists( $lock_file ) ) {
			if ( unlink( $lock_file ) ) {
				$this->exit_ok( "The album has been unlocked" );
			}
			else {
				$this->exit_error( "Cannot unlock album" );
			}
		}
		else {
			$this->exit_ok( "The album was not locked" );
		}		
	}
	
	/**
	 * Render the album edit captions view
	 * @access	private
	 */
	private function _render_album_edit_captions ( ) {
		
		// Must be password protected
		$this->_password_check( $this->params[ "admin_pass" ] );
		
		
		$display_by	= $this->_album_data[ "display_by" ];
		$reverse		= $this->_album_data[ "display_reverse" ];
		
		$image_list	= $this->_album_data[ "sorted" ][ $display_by ];
		if ( $reverse ) { $image_list = array_reverse( $image_list ); }
		
		// What's the empty box value?
		$blank_val	= "(Click to add caption)";
		$url_album	= $this->params["album_url"];
		$album		= $this->_album_name;
		
		$str			= "<a href=\"$url_album\">" . $this->_album_name . "</a>";
		array_push( $this->_html_title, $str );
		array_push( $this->_html_title, "Captions" );

		$link_rotate	= "<a href=\"$url_album&edit=rotate\">Rotate</a>";
		$link_sort		= "<a href=\"$url_album&edit=sort\">Sort</a>";
		$link_caption	= "<a href=\"$url_album&edit=name\">Captions</a>";
		$link_back		= "<a href=\"$url_album\">&lt;&lt; Back</a>";
		array_push( $this->_html_admin, $link_rotate );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_sort );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_caption );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_back );
		
		// If the form has been submitted
		if ( $this->_post["go"] == '1' ) {
			
			// Update the name stored under each image
			foreach ( $image_list as $image ) { 
				
				// Convert the image name to a form-safe name
				$key		= $this->parse_safe_name( $image );
				
				// Get the new value
				$val		= stripslashes( $this->_post[ $key ] );
				
				// If the value == the blank value, then it should be blank
				if ( $val == $blank_val ) { $val = ""; }
				
				// Save our caption
				$this->_album_data[ $image ]["caption"] = $val;
			}
			
			$str	= "<span class=\"good\">Captions updated</span>";
			array_unshift( $this->_html_nav, $str );
		}
		
		$count	= 0;
		$html		= "<form action=\"$url_album&edit=name\" method=\"POST\">\n"
					. $this->html_hidden( "go", "1" )
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "<table border=\"0\" width=\"100%\">\n"
					. "<tr>\n";
				
		foreach ( $image_list as $image ) {
			
			$image_data	= $this->_image_get( $album, $image, "thumb" );
			$image_info	= $this->_album_data[ $image ];
			$img_html	= $image_data["thumb_html"];
			$safe_name	= $this->parse_safe_name( $image );
			$value		= $image_info["caption"];
			
			if ( $value == '' ) { $value = $blank_val; }
			$js_focus	= "if(this.value=='$blank_val'){this.value='';}";
			$js_blur		= "if(this.value==''){this.value='$blank_val';}";
			$html		.= "<td>"
						. $img_html
						. "<p>"
						. "<input type=\"text\" "
							. "name=\"$safe_name\" "
							. "value=\"$value\" "
							. "onFocus=\"$js_focus\" "
							. "onBlur=\"$js_blur\" "
							. ">"
						. "</p>"
						. "</td>\n";
			
			if ( ( ++$count % $this->params["thumb_per_row"] ) == 0 ) { 
				$html .= "</tr>\n<tr>\n"; 
			}
			
		}
		$html		.= "</tr>\n</table>\n"
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "</form>\n";
					
		$this->_output	= $html;
	}
	
	/**
	 * Render the album edit rotate view
	 * @access	private
	 */
	private function _render_album_edit_rotate ( ) {
		
		// Must be password protected
		$this->_password_check( $this->params[ "admin_pass" ] );
		
		$display_by	= $this->_album_data[ "display_by" ];
		$reverse		= $this->_album_data[ "display_reverse" ];
		
		$image_list	= $this->_album_data[ "sorted" ][ $display_by ];
		if ( $reverse ) { $image_list = array_reverse( $image_list ); }
		
		// What's the empty box value?
		$blank_val	= "(Click to add caption)";
		$url_album	= $this->params["album_url"];
		$album		= $this->_album_name;
		
		$str			= "<a href=\"$url_album\">" . $this->_album_name . "</a>";
		array_push( $this->_html_title, $str );
		array_push( $this->_html_title, "Rotate" );

		$link_rotate	= "<a href=\"$url_album&edit=rotate\">Rotate</a>";
		$link_sort		= "<a href=\"$url_album&edit=sort\">Sort</a>";
		$link_caption	= "<a href=\"$url_album&edit=name\">Captions</a>";
		$link_back		= "<a href=\"$url_album\">&lt;&lt; Back</a>";
		array_push( $this->_html_admin, $link_rotate );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_sort );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_caption );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_back );

		// If the form has been submitted
		if ( $this->_post["go"] == '1' ) {
			
			// Update the name stored under each image
			foreach ( $image_list as $image ) { 
				
				// Convert the image name to a form-safe name
				$key		= $this->parse_safe_name( $image );
				
				// Get the new value
				$val		= stripslashes( $this->_post[ $key ] );
				
				// If the value == the blank value, then it should be blank
				if ( $val == $blank_val ) { $val = ""; }
				
				// Save our caption
				$this->_album_data[ $image ]["rotate"] = $val;
			}
			
			$str	= "<span class=\"good\">Rotations updated</span>";
			array_unshift( $this->_html_nav, $str );
		}
		
		$count	= 0;
		$html		= "<form action=\"$url_album&edit=rotate\" method=\"POST\">\n"
					. $this->html_hidden( "go", "1" )
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "<table border=\"0\" width=\"100%\">\n"
					. "<tr>\n";
		
		// What are the options for rotating these pictures?
		$rot_opts	= array( "90&deg; clockwise"       => '270',
									"No rotation"             => '0',
									"90&deg; anti-clockwise"  => '90',
									"180&deg; anti-clockwose" => '180' );

		foreach ( $image_list as $image ) {

			$image_data	= $this->_image_get( $album, $image, "thumb" );
			$image_info	= $this->_album_data[ $image ];
			$img_html	= $image_data["thumb_html"];
			$safe_name	= $this->parse_safe_name( $image );
			
			$caption		= $image_info["caption"];
			if ( $caption == "" ) { $caption = $image_info["name"]; }

			$value		= $image_data["rotate"];
			if ( $value == "" ) { $value = '0'; }
			
			$html			.= "<td>$img_html\n<p><select name=\"$safe_name\">\n";
			foreach ( $rot_opts as $opt_name => $opt_val ) { 
				$select	= ( $opt_val == $value ) ? " selected" : "";
				$html		.= "<option value=\"$opt_val\"$select>"
								. "$opt_name"
							. "</option>\n";
			}
			$html		.= "</select>\n</p>\n</td>\n";
			
			if ( ( ++$count % $this->params["thumb_per_row"] ) == 0 ) { 
				$html .= "</tr>\n<tr>\n"; 
			}
			
		}
		
		$html		.= "</tr>\n</table>\n"
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "</form>\n";
					
		$this->_output	= $html;
	}
	
	/**
	 * Render the album edit sort view
	 * @access	private
	 */
	private function _render_album_edit_sort ( ) {

		// Must be password protected
		$this->_password_check( $this->params[ "admin_pass" ] );
		
		//  Pull the image list
		$image_list	= $this->_album_data[ "sorted" ][ "custom" ];
		
		// What's the empty box value?
		$url_album	= $this->params["album_url"];
		
		$str			= "<a href=\"$url_album\">" . $this->_album_name . "</a>";
		array_push( $this->_html_title, $str );
		array_push( $this->_html_title, "Sort Order" );

		$link_rotate	= "<a href=\"$url_album&edit=rotate\">Rotate</a>";
		$link_sort		= "<a href=\"$url_album&edit=sort\">Sort</a>";
		$link_caption	= "<a href=\"$url_album&edit=name\">Captions</a>";
		$link_back		= "<a href=\"$url_album\">&lt;&lt; Back</a>";
		array_push( $this->_html_admin, $link_rotate );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_sort );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_caption );
		array_push( $this->_html_admin, $this->params["nav_spacer"] );
		array_push( $this->_html_admin, $link_back );
		
		// What are the valid orders for sorting?
		$order_valid	= $this->params["valid_orders"];

		// We're in action mode
		if ( $this->_post["go"] == '1' ) {
			
			// Provided there are items selected in the list
			if ( is_array( $this->_post["custom"] ) ) {
								
				// The items we're moving
				$move			= $this->_post["custom"];
				$dir			= 1;

				// Moving them up or down?
				if      ( $this->_post["up"]   ) { $dir = -1; }
				else if ( $this->_post["down"] ) { $dir = 1;  }

				// Re-order the list
				$new_list	= $this->list_arrange( $image_list, $move, $dir );
				
				// Drop it in place
				$this->_album_data[ "sorted" ][ "custom" ]	= $new_list;
			}
			
			// Default to date if an invalid type is selected
			$order_by		= "date";
			if ( in_array( $this->_post["sortby"], $order_valid ) ) {
				$order_by	= $this->_post["sortby"];
			}
			
			// Save these details
			$this->_album_data["display_by"]			= $order_by;
			$this->_album_data["display_reverse"]	= $this->_post["reverse"];
			
			// Reload the data
			$image_list	= $this->_album_data[ "sorted" ][ "custom" ];
			
			$str	= "<span class=\"good\">Sort order updated</span>";
			array_unshift( $this->_html_nav, $str );			
		}
		
		// We're in load mode
		else {
			$this->_post["sortby"]	= $this->_album_data["display_by"];
			$this->_post["reverse"]	= $this->_album_data["display_reverse"];
		}
		
		// The visible select list
		$select	= '<select name="custom[]" multiple="multiple" size="20">';
		foreach ( $image_list as $image ) {
			$key			= $image;
			$value		= $image;
			$selected	= "";
			if ( is_array( $this->_post["custom"] ) ) {
				if ( in_array( $image, $this->_post["custom"] ) ) { 
					$selected = "selected"; 
				}
			}
			$select	.= "<option value=\"$value\"$selected>$key</option>\n";
		}
		$select	.= "</select>";
		
		$html	= "<form action=\"$url_album&edit=sort\" method=\"POST\">\n"
				. $this->html_hidden( "go", "1" )

				. "<table class=\"normal\" border=\"0\" width=\"100%\">\n"
				. "<tr>\n"
					. "<td width=\"50%\">\n"
					. "<h3>Sort By</h3>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "date", " Image date" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "file", " Image name" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_radio( "sortby", "custom", " Custom order" ) 
					. "</p>\n"
					. "<p>" 
						. $this->html_checkbox( "reverse", "1", " Reverse order" ) 
					. "</p>\n"
					. "<p>" . $this->html_submit( "save", "Save" ) . "</p>\n"
					. "</td>\n"
					. "<td width=\"50%\">\n"
					. "<h3>Custom Order</h3>\n"
					. "<p>"
						. $select
					. "</p>\n"
					. "<p>Move selected "
						. $this->html_submit( "up", "&uarr;" )
						. " "
						. $this->html_submit( "down", "&darr;" )
					. "</p>\n"
					. "</td>\n"
				. "</tr>\n"
				. "</table>\n"
				. "</form>\n";
		
		$this->_output	= $html;
	}
	
	/**
	 * Render the album rebuild view
	 * @access	private
	 */
	private function _render_album_rebuild ( ) {
		
		// The album URL
		$url_album	= $this->params["album_url"];
		
		$str			= "<a href=\"$url_album\">&lt;&lt; Back</a>";
		array_push( $this->_html_admin, $str );
		
		$str			= "<a href=\"$url_album\">" . $this->_album_name . "</a>";
		array_push( $this->_html_title, $str );
		array_push( $this->_html_title, "Rebuild" );
		
		// Print out a table that will rebuild all the images for us
		$image_list	= $this->_album_data[ "sorted" ][ "file" ];
		
		$html			= "<h3>Rebuilding Previews and Thumbnails</h3>\n"
						. "<p>This may take some time, please be patient.</p>\n"
						. "<table border=\"0\" width=\"100%\" "
							. " class=\"rebuild\">\n"
						. "<tr>"
							. "<th width=\"76%\">Image</th>"
							. "<th width=\"12%\">Thumbnail</th>"
							. "<th width=\"12%\">Preview</th>"
						. "</tr>\n";
						
		foreach ( $image_list as $image ) {
			$rotate		= $this->_album_data[ $image ][ "rotate" ];
			$image_enc	= urlencode( $image );
			$html			.= "<tr>"
							. "<td>$image</td>\n"
							. "<td>"
								. "<img src=\"$url_album"
									. "&image=$image_enc"
									. "&rotate=$rotate"
									. "&rebuild=thumb"
									. "&nocache=" . time() . "\">"
							. "</td>\n<td>"
								. "<img src=\"$url_album"
									. "&image=$image_enc"
									. "&rotate=$rotate"
									. "&rebuild=preview"
									. "&nocache=" . time() . "\">"
							. "</td>\n"
							. "</tr>\n";
		}
		$html		.= "</table>\n";
		
		$this->_output	= $html;
	}

	/**
	 * Rebuild a given image
	 * @access	private
	 */
	private function _render_image_rebuild ( ) {

		// By default we spit the cross out
		$output	= $this->_icon_cross;

		// What type of image are we building?
		$rebuild	= $this->_get["rebuild"];
		
		// Thumb or preview only
		if ( ( $rebuild == "thumb" ) or ( $rebuild == "preview" ) ) {
			
			// The specific image, plus any rotation
			$image		= $this->_get["image"];
			$rotate		= $this->_get["rotate"];
			$album		= $this->_get["album"];
			
			$a				= $album;
			$i				= $image;
			$re			= $rebuild;
			$ro			= $rotate;
			
			// Build the image
			$filename	= $this->_image_generate( $a, $i, $re, $ro );
			
			// If we manage to generate an image
			if ( $filename ) {
				
				// Do we display the built image?
				if ( $this->_get["show"] == "image" ) {
					$output		= base64_encode( join( "", file( $filename ) ) );
				}
				
				// We just display the tick
				else {
					$output		 = $this->_icon_tick;
				}
			}
		}
		
		// Output please
		header( "Content-type: image/jpeg" );
		print base64_decode( $output ); 
		exit();
	}

	/**
	 * Render the album fetch view
	 * @access	private
	 */
	private function _render_album_fetch ( ) {
		$this->exit_error( "Sorry, not available yet" );
	}

	/**
	 * Render the album show view
	 * @param	string	$format	Format to output in
	 * @access	private
	 */
	private function _render_album_view ( $format = "html" ) {
		
		$td_width	= $this->params["td_width"];

		$link_home	= "<a href=\"" . $_SERVER["PHP_SELF"] . "\" "
						. " accesskey=\"b\">"
						. "&lt;&lt; Back"
						. "</a>";

		// JSON format gets all this data
		if ( ( $format == "json" ) or ( $format == "jsonurl" ) ) {
			$data					= array();
			
			$data["album"]		= $this->_album_data;
			$data["images"]	= array();
			
			// Scan each image's data, building the full preview/thumbnail data
			foreach ( $this->_album_data["sorted"]["file"] as $image ) {
				$image_data	= $this->_image_get($this->_album_name,$image,"both");
				$image_meta	= $this->_album_data[ $image ];
				$data["images"][ $image ]					= $image_data;
				$data["images"][ $image ]["metadata"]	= $image_meta;
			}
			
			$this->_output		= $data;
			return true;
		}

		$display_by	= $this->_album_data[ "display_by" ];
		$reverse		= $this->_album_data[ "display_reverse" ];
		
		$image_list	= $this->_album_data["sorted"][ $display_by ];
		if ( $reverse ) { $image_list = array_reverse( $image_list ); }
		
		$size			= count( $image_list );
		$page			= sprintf( "%u", $this->_get["page"] );

		// How many pages?
		$per_page	= $this->params["thumb_per_page"];
		$page_total	= sprintf( "%u", ( $size / $per_page ) );
		
		// If we have exactly this number, there's one more page
		if ( ( $size % $per_page ) > 0 ) { $page_total++; }
		
		// What page are we on?
		$page_cur	= sprintf( "%u", ( ( $page / $size ) + 1 ) );

		// Get the URL to link to the album
		$url_album	= $this->params["album_url"];
		
		// What tranche of images are we looking for?
		$start	= ( $page * $per_page );
		$end		= min( $size, ( $start + $per_page ) );

		// Link back to the album list
		$link_album	= "<a href=\"$url_album\">"
						. $this->_album_name
						. "</a>";
		array_push( $this->_html_title, $link_album );
		
		// Start with the album size
		$pic_count	= "$size pictures";
		array_push( $this->_html_nav, $link_home );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		array_push( $this->_html_nav, $pic_count );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		
		// Are we showing all the images?
		if ( $this->_get["page"] == "all" ) {
			$start		= 0;
			$end			= $size;
			$link_pager	= "<a href=\"$url_album\" accesskey=\"a\">"
							. "Show in pages"
							. "</a>";
			array_push( $this->_html_nav, $link_pager );
			array_push( $this->_html_nav, $this->params["nav_spacer"] );
		}
		
		// Page by page
		else if ( $page_total > 1 ) {
			
			// By default all our links are thus
			$link_prev	= "&lt;&lt;";
			$link_next	= "&gt;&gt;";
			$link_start	= "|&lt;&lt;";
			$link_end	= "&gt;&gt;|";
			$link_all	= false;
		
			$prev		= max( 0, $page - 1 );
			$next		= min( $page_total - 1, $page + 1 );
			$last		= $page_total - 1;
			
			if ( $page > 0 ) {
				$link_prev		= "<a href=\"$url_album&page=$prev\""
										. " accesskey=\"p\">&lt;&lt;</a>";
				$link_start		= "<a href=\"$url_album&page=0\" "
										. " accesskey=\"f\">|&lt;&lt;</a>";
			}
			
			if ( $page < ( $page_total - 1 ) ) {
				$link_next		= "<a href=\"$url_album&page=$next\""
										. " accesskey=\"n\">&gt;&gt;</a>";
				$link_end		= "<a href=\"$url_album&page=$last\" "
										. " accesskey=\"l\">&gt;&gt;|</a>";
			}
		
			$link_all	= "<a "
								. "href=\"$url_album&page=all\" "
								. "accesskey=\"a\""
							. ">"
							. "Show all"
							. "</a>";
								
			// What page are we on?
			$text_place	= "Page " . ( $page + 1 ) . " of $page_total";
			
			// Pack up the list of links
			array_push( $this->_html_nav, $link_start );
			array_push( $this->_html_nav, $link_prev );
			array_push( $this->_html_nav, $link_next );
			array_push( $this->_html_nav, $link_end );
			array_push( $this->_html_nav, $text_place );
			array_push( $this->_html_nav, $this->params["nav_spacer"] );
			array_push( $this->_html_nav, $link_all );
			array_push( $this->_html_nav, $this->params["nav_spacer"] );
		}

		
		// Link to the slideshow player
		$url_album	= $this->params["album_url"];	
		$str	= "<a href=\"$url_album&place=0&slideshow=play\" accesskey=\"s\">"
				. "Play slideshow"
				. "</a>";
		array_push( $this->_html_nav, $str );
		
		// Admin link
		$str	= "<a href=\"$url_album&edit=name\" accesskey=\"e\">Edit</a>";
		array_push( $this->_html_admin, $str );
		
		// Now output our list
		$count	= 0;
		$html		= "<a name=\"album\"></a>\n"
					. "<table border=\"0\" width=\"100%\">\n"
					. "<tr>\n"
					. "";

		// Now scan through this slice of the list
		for ( $i = $start ; $i < $end ; $i++ ) {
			
			// Pull the image out
			$image		= $image_list[ $i ];
			$image_data	= $this->_image_get($this->_album_name,$image,"thumb");
			$image_info	= $this->_album_data[ $image ];
			$views		= sprintf( "%u", $image_info[ "views" ] );

			// Pick a nice caption, trim it to fit in the space available
			$caption		= $image_info["caption"];
			$caplen		= $this->params[ "thumb_caption_len"];
			
			// No caption?  Use the image name
			if ( $caption == "" ) { $caption = $image; }
			
			// A caption exists that may need chopping down
			elseif ( strlen( $caption ) > $caplen ) { 
				$caption = substr( $caption, 0, ( $caplen - 3 ) ) . "...";
			}
			
			// Add the image on
			$html	.= ""
						. "<td align=\"center\" width=\"$td_width%\">"
						. "<a name=\"$image\"></a>"
						. "<div class=\"thumb\">"
							. "<a href=\"$url_album&place=$i\" "
							. "title=\"View $image\">"
							. $image_data["thumb_html"]
							. "<br />"
							. $caption
							. "<br>\n$views view" . ( ( $views == 1 ) ? "" : "s" )
							. "</a>\n"
						. "</div>\n"
					. "</td>\n"
					. "";
					
			if ( ( ++$count % $this->params["thumb_per_row"] ) == 0 ) { 
				$html .= "</tr>\n<tr>\n"; 
			}
			
		}
		$html	.= "</tr>\n</table>\n";
		$this->_output	= $html;
	}

	/**
	 * Render the picture show view
	 * @param	string	$format	Format to output in
	 * @access	private
	 */
	private function _render_image_view ( $format = "html" ) {

		$per_page	= $this->params["thumb_per_page"];

		$display_by	= $this->_album_data[ "display_by" ];
		$reverse		= $this->_album_data[ "display_reverse" ];
		
		$image_list	= $this->_album_data["sorted"][ $display_by ];
		if ( $reverse ) { $image_list = array_reverse( $image_list ); }
		
		$album_size	= count( $image_list );

		$which		= sprintf( "%u", $this->_get["place"] );
		$image		= $image_list[ $which ];
		
		// Work out which page of images this one sits in
		$which_page	= sprintf( "%u", ( $which / $per_page ) );
		if ( ( $which % $per_page ) == 0 ) { 
			//$which_page++;
		}

		//$this->debug( "Per page: $per_page, on page: $which_page of $album_size", 2 );
				
		$image_data	= $this->_image_get($this->_album_name,$image,"preview");
		$image_info	= $this->_album_data[ $image ];
		
		$image_file	= $image_data["preview_file"];
		$image_html	= $image_data["preview_html"];
		$image_date	= $image_info["date"];
		$caption		= $image_info["caption"];
		
		if ( $caption != "" ) { $caption .= "<br />\n"; }
		
		$caption		.= "<span class=\"subcaption\">"
						. date( "jS F Y H:i", $image_date )
						. "</span>";

		// Update the number of views for this image
		// this is post-saved at the end of the render process
		$this->_album_data[ $image ][ "views" ]++;
		
		// The URL to the album
		$url_album	= $this->params["album_url"];

		// Build a link to the image itself
		$url_title 	= $url_album
						. "&page=" . sprintf( "%u", $which / $per_page )
						. "#$image"
						. "";
		$str			= "<a href=\"$url_title\" accesskey=\"b\">"
						. $this->_album_name
						. "</a>";
		array_push( $this->_html_title, $str );
		array_push( $this->_html_title, $image );

		// Where are we in the album?
		$text_place	= ( $which + 1 ) . " of $album_size";

		// Is the slideshow playing at the moment?
		$slide_url	= "$url_album&place=$which&slideshow=play";
		$link_slide	= "<a href=\"$slide_url\" accesskey=\"s\">"
						. "Play slideshow"
						. "</a>";

		// Build our links
		$link_prev	= "&lt;&lt;";
		$link_next	= "&gt;&gt;";
		$link_start	= "|&lt;&lt;";
		$link_end	= "&gt;&gt;|";
		
		$prev			= ( $which - 1 );
		$next			= ( $which + 1 );
		$last			= ( $album_size - 1 );
		
		if ( $which > 0 ) {
			$link_prev		= "<a href=\"$url_album&place=$prev\" "
									. "accesskey=\"p\">&lt;&lt;</a>";
			$link_start		= "<a href=\"$url_album&place=0\" "
									. " accesskey=\"f\">|&lt;&lt;</a>";
		}
		if ( $which < ( count( $image_list ) - 1 ) ) {
			$link_next		= "<a href=\"$url_album&place=$next\" "
									. "accesskey=\"n\">&gt;&gt;</a>";
			$link_end		= "<a href=\"$url_album&place=$last\" "
									.  "accesskey=\"l\">&gt;&gt;|</a>";
		}
		
		// The slideshow can't slide off the end of the list
		else { 
			
			// No more slideshow please
			$this->_in_show	= false;
			
			// Stsart the slideshow from the beginning again
			$link_slide	= "<a href=\"$url_album&place=0&slideshow=play\" "
							. " accesskey=\"s\">"
							. "Play slideshow"
							. "</a>";
		}
		
		// If the slidshow is playing then the link is a stop link
		if ( $this->_in_show ) {
			$link_slide	= "<a href=\"$url_album&place=$which\" accesskey=\"s\">"
							. "Stop slideshow"
							. "</a>";
		}

		$link_list	= "<a href=\"$url_album&page=$which_page\" "
						. " accesskey=\"b\">"
						. "&lt;&lt; Back"
						. "</a>";

		// Link to the full size version
		$url_fs		= $image_data["src"];
		$link_fs		= "<a href=\"$url_fs\" accesskey=\"v\">"
						. "View full size"
						. "</a>";
						
		array_push( $this->_html_nav, $link_list );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		array_push( $this->_html_nav, $link_start );
		array_push( $this->_html_nav, $link_prev );
		array_push( $this->_html_nav, $link_next );
		array_push( $this->_html_nav, $link_end );
		array_push( $this->_html_nav, $text_place );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		array_push( $this->_html_nav, $link_slide );
		array_push( $this->_html_nav, $this->params["nav_spacer"] );
		array_push( $this->_html_nav, $link_fs );

		// Admin link
		$str	= "<a href=\"$url_album&edit=name\" accesskey=\"e\">Edit</a>";
		array_push( $this->_html_admin, $str );
		
		// Link to the fullsize version of the image, and image click
		$img_html		= "<a href=\"$url_fs\" accesskey=\"v\">$image_html</a>";

		$exif_details	= array();
		$exif_str		= "";
		$image_meta 	= $image_info[ "metadata" ];
		
		// Exif details
		if ( $image_meta[ "camera_model" ] != "" ) {
			$str	= "" . $image_meta[ "camera_model" ];
			array_push( $exif_details, $str );
		}
		
		if ( $image_meta["exposure"] != "" ) {
			$str	= "" . $image_meta[ "exposure" ] . "s";
			array_push( $exif_details, $str );
		}

		if ( $image_meta["fnumber"] != "" ) {
			$str	= "f/" . $image_meta[ "fnumber" ];
			array_push( $exif_details, $str );
		}

		if ( $image_meta["iso"] != "" ) {
			$str	= "ISO" . $image_meta[ "iso" ];
			array_push( $exif_details, $str );
		}

		if ( count( $exif_details ) > 0 ) {
			$exif_str	= "<p class=\"subcaption\">"
							. join( ", ", $exif_details )
							. "</p>\n";
		}

		// Finally output the image
		$html	= "<div class=\"preview\">\n"
					. $img_html 
					. "<p class=\"caption\">$caption</p>\n"
					. $exif_str
				. "</div>\n";
				
		$this->_output	= $html;
	}

	/**
	 * Check that the current configuration is valid
	 * @return	boolean	True if all the configuration settings are OK
	 * @access	private
	 */
	private function _config_check ( ) {
		
		// These are the functions that are required by the code
		$func_list	= array( 
							"imagecreatefromstring",
							"imagecreatetruecolor",
							"imagecopyresized",
							"imagecopyresampled",
							"imagejpeg",
							"imagerotate",
							"imagedestroy",
							"getimagesize",
						);
		
		// Scan for each of these functions
		foreach ( $func_list as $func_name ) {
			$func_error	= "The required function '$func_name' is not installed "
							. "or available.  Please check your PHP server "
							. "configuration.";
			if ( !( function_exists( $func_name ) ) ) {
				$this->exit_error( $func_error );
				return false;
			}
		}
		
		if ( !( file_exists( $this->_folder_base ) ) ) {
			$this->exit_error( "The 'base' folder does not exist" );
			return false;
		}
		
		if ( !( is_readable( $this->_folder_base ) ) ) {
			$this->exit_error( "Unable to read from the 'base' folder" );
			return false;
		}

		if ( !( file_exists( $this->_folder_gen ) ) ) {
			$this->exit_error( "The 'generated' folder does not exist" );
			return false;
		}
		
		if ( !( is_readable( $this->_folder_gen ) ) ) {
			$this->exit_error( "Unable to read from the 'generated' folder" );
			return false;
		}
		
		if ( !( is_writable( $this->_folder_gen ) ) ) {
			$this->exit_error( "Unable to write to the 'generated' folder" );
			return false;
		}
		
		return true;
	}
	
	/*
	 * Check that we have a password and it is the correct one
	 * @param	string	$pass		Password to check for
	 * @return 	boolean				True if the password is correct
	 * @access	private
	 */
	private function _password_check ( $pass ) {
		if ( $_SERVER["PHP_AUTH_PW"] != $pass ) {
			header( 'WWW-Authenticate: Basic realm="MicroAlbum Admin"' );
			header( 'HTTP/1.0 401 Unauthorized' );
			exit_error( "Incorrect password" );
			return false;
		}
		else {
			return true;
		}
	}
	
	/*
	 * Check that an album exists and exit if there's an error
	 * @param	string	$album	The name of the album folder
	 * @return 	boolean				True if the album was found
	 * @access	private
	 */
	private function _album_exists ( $album ) {
		$folder			= $this->_folder_base . "/" . $album;
		$dirname			= dirname( $folder );

		// Dirname must match the album name
		if ( $dirname != $this->_folder_base ) { 
			$this->exit_error( "The album is not in the 'base' folder" ); 
			return false;
		}

		// Album directory must exist
		if ( !( is_dir( $album ) ) ) { 
			$this->exit_error( "This album does not exist" );
			return false;
		}	
		
		return true;
	}

	/**
	 * Return a list of all the albums that exist
	 * @return 	array 				The list of albums
	 * @access	private
	 */
	private function _album_list ( ) {
		
		$invalid	= array( ".", "..", $this->_folder_gen );
		$list		= array();
		$dh		= opendir( $this->_folder_base );
		
		while ( $file = readdir( $dh ) ) {
			$the_file	= $this->_folder_base . "/" . $file;
			if ( is_dir( $the_file ) and
				( !( in_array( $file, $invalid ) ) ) ) {
				$tmp				= array();
				$tmp["file"]	= $file;
				$tmp["mtime"]	= filemtime( $the_file );
				$tmp["etime"]	= 0;
				array_push( $list, $tmp );
			}
		}

		// Sort by the date on the folder
		usort( $list, array( "MicroAlbum", "_cmp_mtime" ) );
		$return 	= array();

		// Strip off the folder name from the front
		foreach ( array_reverse( $list ) as $chaff => $data ) {
			array_push( $return, $data["file"] );
		}
		
		return $return;
	}
	
	/**
	 * Load the album summary data file
	 * @access	private
	 */
	private function _album_summary_load ( ) {
		
		$data_file	= $this->_folder_base . "/"
						. $this->_folder_gen . "/"
						. "album_list.dat";
		$data_file	= realpath( $data_file );

		$return 		= array();
		$changed		= false;
		$album_list	= $this->_album_list( $base_folder );

		// If the data file exists, get it.
		if ( file_exists( $data_file ) ) {
			$return 	= unserialize( join( "", file( $data_file ) ) );
		}

		// If the custom order isn't present, take the list of albums
		if ( !( is_array( $return["custom_order"] ) ) ) { 
			$return["custom_order"]	= $album_list;
			$changed						= true;
			}

		// By default, we'll display by folder name
		if ( !( $return["display_by"] ) ) {
			$return["display_by"]	= "file";
			$changed						= true;
		}

		// Check that the current list of albums doesn't contain anything more
		// than what's in the custom list
		$removed	= array_diff( $return["custom_order"], $album_list );
		$added	= array_diff( $album_list, $return["custom_order"] );

		// Remove any albums that no longer exist
		foreach ( $removed as $album ) {
			$changed	= true;
			foreach ( array_keys( $return["custom_order"], $album ) as $key ) {
				unset( $return["custom_order"][ $key ] );
			}
		}

		// Add any albums that have been added to the start of the list
		foreach ( $added as $album ) {
			$changed	= true;
			array_unshift( $return["custom_order"], $album );
		}

		// Save it if we've made some changes
		if ( $changed ) { $this->_album_summary_save( $return ); }
		return $return;
	}
	
	/**
	 * Save the album summary data to file
	 * @param	mixed		$data		Nested array of summary data
	 * @access	private
	 */
	private function _album_summary_save ( $data ) {
		$data_file	= $this->_folder_base . "/"
						. $this->_folder_gen . "/"
						. "album_list.dat";
		return $this->_serialize_to_file( $data_file, $data );
	}
	
	/**
	 * Pick one random image from the given album
	 * @param	string	$album	Name of the album
	 * @param	string	$prefix	Image type: thumb or preview
	 * @return 	mixed					The image data returne from _image_get()
	 * @access	private
	 */
	private function _album_get_random ( $album, $prefix = "thumb" ) {
		
		// Pull up the list of pictures from the album data
		$album_data	= $this->_album_data_load( $album );
		$file_list	= $album_data["sorted"]["file"];
		
		// Which random image do we want?
		$count		= count( $file_list );
		$which		= rand( 1, $count );
		$which--;
		
		// Pick the name of an image
		$image		= $file_list[ $which ];
		
		// Return the relevant data
		return $this->_image_get( $album, $image, $prefix );
	}
	
	/**
	 * Get the saved data file for the given album
	 * @param	string	$album	The album to load data for
	 * @return 	mixed					The album data as a nested array
	 * @access	private
	 */
	private function _album_data_load ( $album ) {
		
		if ( !( $this->_album_exists( $album ) ) ) { 
			$this->exit_error( "This album does not exist" );
			return false;
		}
		
		$data_folder	= $this->_folder_base . "/" 
							. $this->_folder_gen . "/"
							. $album;
		$data_file		= $data_folder . "/" . "album.dat";
		
		// Create the data folder, if required
		if ( !( is_dir( $data_folder ) ) ) {
			if ( !( mkdir( $data_folder, 0755 ) ) ) {
				return $this->exit_error( "Unable to create album cache folder" );
			}
		}
		
		// Rebuild the album data
		$album_data		= array();
		
		if ( file_exists( $data_file ) ) {
			$album_data 	= unserialize( join( "", file( $data_file ) ) );
		}
		else {
			$album_data 	= $this->_album_data_rebuild( $album );
		}
		
		// Get the mtime on the album itself
		$album_folder	= $this->_folder_base . "/" . $album;
		$album_mtime	= filemtime( $album_folder );
		
		// If the time doesn't match, rebuild the album data
		if ( $album_data["album_mtime"] != $album_mtime ) {
			$new_data		= $this->_album_data_rebuild( $album, $album_data );
			$album_data		= $new_data;
		}
		
		return $album_data;
	}

	/**
	 * Rebuild the data file for an album
	 * @param	string	$album	The album to load data for
	 * @param	mixed		$source	Optional original data to work from
	 * @return 	mixed					The rebuilt album data as a nested array
	 * @access	private
	 */
	private function _album_data_rebuild ( $album, $source = false ) {
		
		$this->debug( "Rebuilding data for: $album", 2 );
		
		// The default album data
		$album_data		= array();
		
		// If there is source data, take it.
		if ( is_array( $source ) ) { $album_data = $source; }

		// Set defaults, if none are present
		if ( !( is_array( $album_data["sorted"] ) ) ) {
			$album_data["sorted"]				= array();
			$album_data["sorted"]["file"]		= array();
			$album_data["sorted"]["date"]		= array();
			$album_data["sorted"]["custom"]	= array();
		}
		
		// Set the default display order
		if ( $album_data["display_by"] == "" ) {
			$album_data["display_by"]			= "file";
		}
		
		// Set the default reversing preference
		if ( $album_data["display_reverse"] == "" ) {
			$album_data["display_reverse"]	= false;
		}

		// Now build a full list of filenames
		$album_path		= $this->_folder_base . "/" . $album;
		$album_path		= realpath( $album_path );
		$album_mtime	= filemtime( $album_path );
		
		// Store the mtime for this album
		$album_data["album_mtime"]	= $album_mtime;

		// Open up the album directory
		$dh				= opendir( $album_path );
		$sorted_file	= array();
		
		// Read the album in
		while ( $file = readdir( $dh ) ) {
			
			$the_file	= realpath( "$album_path/$file" );
			
			if ( is_file( $the_file ) and
				( $file != '.' ) and
				( $file != ".." ) and
				( eregi( "^$album_path/", $the_file ) ) ) {
				
				// Check it's a valid file type
				$parts	= pathinfo( $the_file );
				$ext		= strtolower( $parts[ "extension" ] );
				if ( in_array( $ext, $this->params["valid_types"] ) ) {
					array_push( $sorted_file, $file );
				}
			}
		}

		// Sort by filename and re-order the key values
		natcasesort( $sorted_file );
		$re_sorted		= array_values( $sorted_file );
		$sorted_file	= $re_sorted;
		
		$sorted_pre		= array();
		foreach ( $sorted_file as $image ) {
			
			// Build a small item for this image
			$item				= array();
			$item["path"]	= $full_path;
			$item["image"]	= $image;

			// Full path to the file
			$full_path		= $album_path .  "/" . $image;

			// If there is stored album data, take it.
			if ( $album_data[ $image ]["date"] != "" ) {
				$item["mtime"]	= $album_data[ $image ]["date"];
			}
			
			// Album data is not stored, try calculating it
			else {
				
				// Read the exif data
				$exif			= @exif_read_data( $full_path );
				$the_time	= 0;
				
				// Try the original image date
				if ( $exif["DateTimeOriginal"] != "" ) {
					$date							= $exif["DateTimeOriginal"];
					list( $dt, $tm )			= explode( " ", $date );
					list( $yr, $mo, $da )	= explode( ":", $dt );
					$date							= "$yr-$mo-$da $tm";
					$the_time					= strtotime( $date );
				}

				// Try the actual DateTime
				else if ( $exif["DateTime"] != "" ) {
					$date							= $exif["DateTime"];
					list( $dt, $tm )			= explode( " ", $date );
					list( $yr, $mo, $da )	= explode( ":", $dt );
					$date							= "$yr-$mo-$da $tm";
					$the_time					= strtotime( $date );
				}

				// Try the timestamp
				else if ( $exif["FileDateTime"] != "" ) {
					$the_time					= $exif["FileDateTime"];
				}
				
				// Take the file mtime by default
				else {
					$the_time					= filemtime( $full_path );
				}
				
				$item["mtime"]	= $the_time;
			}
			
			// Add to the list
			array_push( $sorted_pre, $item );
		}
		
		// Sort the list by date properly
		usort( $sorted_pre, array( "MicroAlbum", "_cmp_mtime" ) );
		
		// Pull out the image names
		$sorted_date	= array();
		foreach ( $sorted_pre as $item ) {
			array_push( $sorted_date, $item["image"] );
		}
		
		// Take the provisional custom sorted list
		$sorted_custom	= $album_data["sorted"]["custom"];
		if ( !( is_array( $sorted_custom ) ) ) { 
			$sorted_custom	= array();
		}
		
		// Work out what's gone and what's new
		$custom_del		= array_diff( $sorted_custom, $sorted_file );
		$custom_add		= array_diff( $sorted_file, $sorted_custom );
		
		// Rebuild this to a new custom list
		$custom_new		= array();
		
		// First, remove any images that have gone
		foreach ( $sorted_custom as $image ) {
			if ( in_array( $image, $custom_del ) ) { 
				$this->debug( "Removed $image from $album", 2 );
			}
			else {
				array_push( $custom_new, $image );
			}
		}
		
		// Then add on any new images
		foreach ( $custom_add as $image ) {
			$this->debug( "Added $image to $album", 2 );
			array_push( $custom_new, $image );
		}
		
		$sorted_custom	= $custom_new;
		
		// Save the sorting data
		$album_data["sorted"]["file"]		= $sorted_file;
		$album_data["sorted"]["date"]		= $sorted_date;
		$album_data["sorted"]["custom"]	= $sorted_custom;
		
		// Save it all
		$this->_album_data_save( $album, $album_data );
		
		return $album_data;
	}

	/**
	 * Write the album data out to it's data file
	 * @param	string	$album	Album name
	 * @param	mixed		$data		Nested array of album data
	 * @access	private
	 */
	private function _album_data_save ( $album, $data ) {
		$data_file	= $this->_folder_base . "/" 
						. $this->_folder_gen . "/"
						. $album . "/"
						. "album.dat";
		return $this->_serialize_to_file( $data_file, $data );
	}

	/**
	 * Serialize the given data to the given file
	 * @param	string	$file		File to write to
	 * @param	string	$data		Data to write
	 * @return 	boolean				True on success
	 * @access 	private
	 */
	private function _serialize_to_file( $file, $data ) { 

		// What's the lock file?
		$lock_file		= "$file.lock";
		$wait				= file_exists( $lock_file );
		$loop_count		= 0;

		while( $wait ) {

			usleep( rand( 10000, 100000 ) );

			$wait				= file_exists( $lock_file );
			$loop_count++;

			if ( $loop_count > 20 ) { 

				// Remember the current album
				$php_self	= $_SERVER["PHP_SELF"];
				$album_enc	= rawurlencode( $params["cur_album"] );
				$url_album	= "$php_self?album=$album_enc";

				$error	= "The album is locked, the data file cannot be saved. "
							. "Try loading the page again, see if the error clears. "
							. "If the problem persists "
							. "<a href=\"$url_album&unlock=yes\">"
							. "click here to try to unlock the album."
							. "</a>"
							. "";
				exit_error( "$error" );
			}
		}

		// Write the lockfile
		touch( $lock_file );

		$data_file		= $file;
		$data_data		= serialize( $data );
		$fh				= fopen( $data_file, "w" );
		fwrite( $fh, $data_data );
		fclose( $fh );

		// Save data
		$this->debug( "Written data to $data_file", 2 );

		// Remove the lock file
		unlink( $lock_file );

		return true;
	}
	
	/**
	 * Get all the data about an image, including dimensions,
	 * size, name, date, thumbnail, preview, etc.
	 * @param	string	$album		The album folder
	 * @param	string	$image		The image in that folder
	 * @param	string	$get_type	Sub-type to fetch: preview, thumb, both
	 * @return	mixed		$return		Data about the image
	 * @access 	private
	 */
	private function _image_get ( $album, $image, $get_type = "both" ) {
		
		$album_data	= false;
		$image_data	= false;
		
		// Load the data file
		if ( $this->_in_album ) {
			$image_data	= $this->_album_data[ $image ];
		}
		else {
			$album_data	= $this->_album_data_load( $album );
			$image_data	= $album_data[ $image ];
		}
		
		// The actual image file itself
		$the_image			= $this->_folder_base . "/" . $album . "/" . $image;

		// The actual file
		$return 				= array();
		$return["file"]	= realpath( $the_image );

		// If the data format version does not match
		if ( $image_data[ "version" ] != $this->params["data_version"] ) {
			
			// Update the version code on the data format
			$image_data[ "version" ]	= $this->params["data_version"];

			// Take the exif metadata
			$exif								= @exif_read_data( $return["file"] );
			$meta 							= array();
			$meta["camera_make"]			= $exif["Make"];
			$meta["camera_model"]		= $exif["Model"];
			$meta["exposure"]				= $exif["ExposureTime"];
			$meta["fnumber"]				= $exif["FNumber"];
			$meta["iso"]					= $exif["ISOSpeedRatings"];
			$meta["digitized"]			= $exif["DateTimeDigitized"];
			$meta["DateTime"]				= $exif["DateTime"];
			$meta["DateTimeOriginal"]	= $exif["DateTimeOriginal"];

			// Format the f number better, if it is in two parts
			if ( $meta["fnumber"] ) {
				list( $left, $right )	= explode( "/", $meta["fnumber"] );
				if ( ( $left != "" ) && ( $right != "" ) ) {
					$meta["fnumber"]		= sprintf( "%.1f", $left / $right );
				}
			}
			
			// Format the exposure time better
			if ( $meta["exposure"] ) {
				list( $left, $right )	= explode( "/", $meta["exposure"] );
				if ( ( $left != "" ) && ( $right != "" ) ) {
					$meta["exposure"]		= "1/"
												. sprintf( "%u", ( $right / $left ) );
				}
			}

			// Remove the camera name from the start of the model
			$cam_make			= $meta["camera_make"];
			$cam_model			= $meta["camera_model"];
			$camera				= eregi_replace( "^$cam_make", "", $cam_model );
			$meta["camera"]	= $cam_make . $camera;

			// Clean the metadata of excess spaces
			foreach ( $meta as $key => $val ) { $meta[ $key ] = trim( $val ); }

			// Store the image metadata
			$image_data[ "metadata" ]	= $meta;

			$the_time	= time();

			// Try the original image date
			if ( $exif["DateTimeOriginal"] != "" ) {
				$date							= $exif["DateTimeOriginal"];
				list( $dt, $tm )			= explode( " ", $date );
				list( $yr, $mo, $da )	= explode( ":", $dt );
				$date							= "$yr-$mo-$da $tm";
				$the_time					= strtotime( $date );
			}
			
			// Try the actual DateTime
			else if ( $exif["DateTime"] != "" ) {
				$date							= $exif["DateTime"];
				list( $dt, $tm )			= explode( " ", $date );
				list( $yr, $mo, $da )	= explode( ":", $dt );
				$date							= "$yr-$mo-$da $tm";
				$the_time					= strtotime( $date );
			}
			
			// Try the timestamp
			else if ( $exif["FileDateTime"] != "" ) {
				$the_time					= $exif["FileDateTime"];
			}
			
			// Pick the file mtime by default
			else {
				$the_time					= filemtime( $return["file"] );
			}

			// Pick a suitable date for it
			$image_data[ "date" ]		= $the_time;

			// Specify the image dimensions
			$img_size						= getimagesize( $return["file"] );
			$image_data["x"]				= $img_size[ 0 ];
			$image_data["y"]				= $img_size[ 1 ];
			
			// If we're not in an album then we should save the data
			if ( $this->_in_album ) {
				$this->_album_data[ $image ]	= $image_data;
			}
			else {
				$album_data[ $image ]			= $image_data;
				$this->_album_data_save( $album, $album_data );
			}
			
			$this->debug( "_image_get: Rebuilt data for $album/$image", 2 );
		}

		// The actual file
		$return["src"]				= $this->_image_src( $album, $image );
		$return["name"]			= $image;

		// Possible image rotation
		$return["rotate"]			= $image_data["rotate"];
		$rot							= $return["rotate"];

		// Get the file modification time
		$return["file_date"]		= filemtime( $return["file"] );
		$return["date"]			= $image_data["date"];

		// Source image X and Y dimensions
		$sx							= $image_data["x"];
		$sy							= $image_data["y"];

		$return["x"]				= $sx;
		$return["y"]				= $sy;
		$return["html"]			= "<img "
										. "src=\""    . $return["src"] . "\" "
										. "width=\""  . $return["x"] . "\" "
										. "height=\"" . $return["y"] . "\" "
										. "alt=\""    . $image . "\""
										. ">";

		$get_thumb					= false;
		$get_preview				= false;
		
		$this->debug( "_image_get: Fetch $get_type for $image", 2 );
		
		// Which of these are we to retrieve?
		if ( $get_type == "thumb" ) {
			$get_thumb					= true;
		}
		else if ( $get_type == "preview" ) {
			$get_preview				= true;
		}
		else {
			$get_thumb					= true;
			$get_preview				= true;
		}

		// Get the thumb?
		if ( $get_thumb == true ) {
			$thumb	= $this->_image_get_generated( $album,
							$image,
							$sx,$sy,
							"thumb",
							$rot
						);
			$return["thumb_file"]	= $thumb["file"];
			$return["thumb_src"]		= $thumb["src"];
			$return["thumb_x"]		= $thumb["x"];
			$return["thumb_y"]		= $thumb["y"];
			$return["thumb_html"]	= "<img "
											. "src=\""    . $return["thumb_src"] . "\" "
											. "width=\""  . $return["thumb_x"] . "\" "
											. "height=\"" . $return["thumb_y"] . "\" "
											. "alt=\"Thumbnail of " . $image . "\""
											. ">";
		}

		// Get the preview?
		if ( $get_preview == true ) {
			$preview	= $this->_image_get_generated( $album,
							$image,
							$sx,$sy,
							"preview",
							$rot
						);
			$return["preview_file"]	= $preview["file"];
			$return["preview_src"]	= $preview["src"];
			$return["preview_x"]		= $preview["x"];
			$return["preview_y"]		= $preview["y"];
			$return["preview_html"]	= "<img "
											. "src=\""    . $return["preview_src"] . "\" "
											. "width=\""  . $return["preview_x"] . "\" "
											. "height=\"" . $return["preview_y"] . "\" "
											. "alt=\"Preview of " . $image . "\""
											. ">";
		}
		return $return;		
	}
	
	/**
	 * Get the html 'src' of an image, suitable for an <img> tag
	 * @param	string	$album	The album folder
	 * @param	string	$image	The image in that folder
	 * @return 	string			 	String suitable for <img> tag
	 * @access	public
	 */
	private function _image_src ( $album, $image ) {
		$parts	= array_map( "rawurlencode", explode( "/", $album ) );
		return join( "/", $parts ) . "/" . rawurlencode( $image );
	}
	
	/*
	 * Return data for the given image in the given album, including
	 * data such as the filename, dimensions and src string
	 * @param	string	$album	The album folder
	 * @param	string	$image	The image in that folder
	 * @param	integer	$sx		Source image X dimension
	 * @param	integer	$sy		Source image Y dimension
	 * @param	string	$prefix	Optional prefix to apply (thumb or preview)
	 * @param	integer	$rot		Optional rotation to apply in degrees
	 * @return 	array				 	Data about the given image
	 * @access	public
	 */
	private function _image_get_generated ( $album, $image, $sx, $sy, $prefix = "thumb", $rot = 0 ) {

		if ( !( $prefix != "thumb" ) or ( $prefix != "preview" ) ) {
			$prefix	= "thumb";
		}

		$base				= $this->_folder_base;
		$gen_folder		= $this->_folder_gen;
		$target_x		= $this->params[ $prefix . "_x" ];
		$target_y		= $this->params[ $prefix . "_y" ];
		$src_file		= realpath( "$base/$album/$image" );

		// Get the source dimensions as an array
		$src_dim			= array();
		$src_dim[ 0 ]	= $sx;
		$src_dim[ 1 ]	= $sy;

		$src_x			= $sx;
		$src_y			= $sy;

		// Get the original non-rotated constrained dimensions
		$orig_dim	= $this->_image_constrain( $src_x, $src_y, $target_x, $target_y );
		$orig_x		= sprintf( "%u", $orig_dim['x'] );
		$orig_y		= sprintf( "%u", $orig_dim['y'] );

		// 90 or 180 degrees of rotation?  The dimensions are reversed
		if ( ( $rot == '90' ) or ( $rot == '270' ) ) {
			$src_x	= $src_dim[ 1 ];
			$src_y	= $src_dim[ 0 ];
			$rot_str	= "r$rot-";
		}
		else if ( $rot == '180' ) { 
			$rot_str	= "r$rot-";
		}

		// Constrain it to fit in the box
		$dim				= $this->_image_constrain( $src_x, $src_y, $target_x, $target_y );
		$dst_x			= sprintf( "%u", $dim['x'] );
		$dst_y			= sprintf( "%u", $dim['y'] );

		// All built images are in jpg format
		$parts			= pathinfo( $image );
		$ext				= strtolower( $parts[ "extension" ] );
		$dest_image		= eregi_replace( ".$ext$", ".jpg", $image );

		// Make sure that we know what file and where it's being built
		$gen_dir			= realpath( "$base/$gen_folder" );
		$dest_dir		= realpath( "$base/$gen_folder/$album" );	
		
		// This is the final file we're looking for, possibly rotated
		$dest_file		= "$prefix-$dst_x-$dst_y-$rot_str$dest_image";

		// This is the substitue file, if the rotated version doesn't exist
		$dest_file_nr	= "$prefix-$orig_x-$orig_y-$dest_image";

		$the_file		= "$dest_dir/$dest_file";
		$img_src			= $this->_folder_gen . "/$album";
		$the_src			= $this->_image_src( $img_src, $dest_file );

		// If the (transformed) file doesn't exist, we may need to generate it
		if ( !( file_exists( "$dest_dir/$dest_file" ) ) ) {

			$do_rebuild		= true;

			// The final file doesn't exist.  Do we live-build the version?
			if ( $this->params["live_build"] == true ) {
				$the_file	= $src_file;
				$the_src		= "index.php/" . rawurlencode( $dest_file )
								. "?album=" . rawurlencode( $album )
								. "&image=" . rawurlencode( $image )
								. "&rotate=$rot"
								. "&rebuild=$prefix"
								. "&show=image"
								. "&nocache=" . time()
								. "";
				$do_rebuild	= false;
				$this->debug( "Live build of $prefix: $album/$image", 2 );
			}
			
			// No live-build.  What about the non-rotated version?
			else if ( file_exists( "$dest_dir/$dest_file_nr" ) ) {
				$the_file	= "$dest_dir/$dest_file_nr";
				$img_src		= $this->_folder_gen . "/$album";
				$the_src		= $this->_image_src($img_src,$dest_file_nr);
				$this->debug( "Rebuild of $prefix needed: $album/$image", 2 );
			}

			// No original image, no non-rotated version, use the source image
			else {
				$the_file	= $src_file;
				$the_src		= $this->_image_src( $album, $image );
				$this->debug( "Rebuild of $prefix needed: $album/$image", 2 );
			}
			
			// No live-build?  No image?  We must flag an album rebuild.
			if ( $do_rebuild == true ) {
				$this->params["album_rebuild"][ $album ]	= true;
			}
			
		}

		$return 				= array();
		$return["file"]	= $the_file;	// Full path to the file
		$return["src"]		= $the_src;		// Relative path to the file
		$return["x"]		= $dst_x;		// x of image
		$return["y"]		= $dst_y;		// y of image
		
		return $return;
	}

	/**
	 * Generate a preview or thumbnail image
	 * @param	string	$album	Name of the album
	 * @param	string	$image	Name of the image
	 * @param	string	$prefix	Optional type, thumb or preview
	 * @param	integer	$rot		Optional rotation, in degrees
	 * @return 	string				The filename of the built image
	 * @access	private
	 */
	private function _image_generate ( $album, $image, $prefix = "thumb", $rot = 0 ) {

		// Valid album and prefix type are required
		if ( ( $prefix != "thumb" ) and ( $prefix != "preview" ) ) { 
			$prefix	= "thumb";
		}
		
		// The folders we work from
		$base_dir	= $this->_folder_base;
		$gen_dir		= $this->_folder_gen;
		
		// The maximum dimensions the generated image can fit into
		$tgt_x		= $this->params[ $prefix . "_x" ];
		$tgt_y		= $this->params[ $prefix . "_y" ];
		$tgt_qual	= $this->params[ $prefix . "_quality" ];
		
		// The source image file
		$src_file	= realpath( "$base_dir/$album/$image" );
		
		// File doesn't exist?
		if ( !( $src_file ) ) { return false; }

		// File isn't inside the base folder?
		if ( !( eregi( "^$base_dir/", $src_file ) ) ) { return false; }

		$tgt_dir		= realpath( "$base_dir/$gen_dir/$album" );
		
		// Invalid path?
		if ( $tgt_dir == "" ) { return false; }
		
		// If the target does doesn't exist, try making it.
		if ( !( is_dir( $tgt_dir ) ) ) {
			if ( !( mkdir( $dest_dir, 0755 ) ) ) { return false; }
		}
		
		// This folder must now be writable
		if ( !( is_writable( $tgt_dir ) ) ) { return false; }
		
		// We build all the images as JPG files.  Rename the file to 
		// have the .jpg extension
		$parts		= pathinfo( $image );
		$ext			= strtolower( $parts[ "extension" ] );
		$dest_image	= eregi_replace( ".$ext$", ".jpg", $image );
		
		// Is there any rotation applied to the file?
		$rot_str		= "";

		// Load up the source image
		$src_size	= getimagesize( $src_file );
		$do_rot		= false;

		// The source image dimensions
		$src_x		= $src_size[ 0 ];
		$src_y		= $src_size[ 1 ];

		$orig_x		= $src_size[ 0 ];
		$orig_y		= $src_size[ 1 ];

		// What is the size of an un-rotated image?
		$small_size	= $this->_image_constrain( $src_x, $src_y, $tgt_x, $tgt_y );
		$small_x		= $small_size[ 0 ];
		$small_y		= $small_size[ 1 ];

		// Only these levels of rotation are allowed
		if ( ( $rot == '90' ) or ( $rot == '270' ) ) {
			
			$do_rot	= true;
			$rot_str	= "r$rot-";
			
			// The X and Y are switched
			$src_x	= $src_size[ 1 ];
			$src_y	= $src_size[ 0 ];
		}
		
		// Straight flip
		else if ( $rot == "180" ) {
			$do_rot	= true;
			$rot_str	= "r$rot-";
			
			// The X and Y are not flipped
			$src_x	= $src_size[ 0 ];
			$src_y	= $src_size[ 1 ];
		}

		// The size of the final image, possibly post-rotation
		$dst_size	= $this->_image_constrain( $src_x, $src_y, $tgt_x, $tgt_y );
		$dst_x		= sprintf( "%u", $dst_size[ 0 ] );
		$dst_y		= sprintf( "%u", $dst_size[ 1 ] );
		
		// The full generated filename, encoding size, rotation and name
		$tgt_file	= "$prefix-$dst_x-$dst_y-$rot_str$dest_image";
		$full_path	= "$tgt_dir/$tgt_file";

		// If the file's not there, build it.
		if ( !( file_exists( $full_path ) ) ) {

			// Load the source image
			$src	= imagecreatefromstring( join( "", file( $src_file ) ) );
			
			// Create the destionation image
			$dst	= imagecreatetruecolor( $small_x, $small_y );
			
			// Copy the source image to the destionation image, unrotated
			$this->fastimagecopyresampled( $dst, $src, 
												  0, 0, 
												  0, 0,
												  $small_x, $small_y, 
												  $orig_x, $orig_y,
												  $tgt_qual );

			// Do we need to rotate the image?  If so we will use the
			// new, smaller image to do it.
			if ( $do_rot ) {
				$tmp	= imagerotate( $dst, $rot, 0 );
				$dst	= $tmp;
				unset( $tmp );
			}

			// Save to JPEG?
			if ( imagejpeg( $dst, $full_path, 90 ) ) { 
				imagedestroy( $src );
				imagedestroy( $dst );
			}
			else {
				imagedestroy( $src );
				imagedestroy( $dst );
				return false;
			}
		}
		
		// Return the pathname of the file
		return $full_path;
	}

	/**
	 * Constrain an image to fit inside a given box, maintaining the
	 * original aspect ratio of the source dimensions
	 * @param	integer	$img_x	The width of the original image
	 * @param	integer	$img_y	The height of the original image
	 * @param	integer	$to_x		The width of the constraint box
	 * @param	integer	$to_y		The height of the constraint box
	 * @return	mixed					The new X/Y of the image
	 * @access	public
	 */
	private function _image_constrain ( $img_x, $img_y, $to_x, $to_y ) {
		$new_x		= $img_x;
		$new_y		= $img_y;
		
		if ( ( $img_x > $to_x ) or ( $img_y > $to_y ) ) {
			if ( $img_x >= $img_y ) {
				$new_x	= $to_x;
				$new_y	= $new_x * ( $img_y / $img_x );
				if ( $new_y > $to_y ) {
					$new_y	= $to_y;
					$new_x	= $new_y * ( $img_x / $img_y );
				}
			}
			else {
				$new_y	= $to_y;
				$new_x	= $new_y * ( $img_x / $img_y );
				if ( $new_x > $to_x ) {
					$new_x	= $to_x;
					$new_y	= $new_x * ( $img_y / $img_x );
				}
			}
		}
		
		$new_x		= sprintf( "%u", $new_x );
		$new_y		= sprintf( "%u", $new_y );
		
		$ret 			= array();
		$ret["x"]	= $new_x;
		$ret["y"]	= $new_y;
		$ret[ 0 ]	= $new_x;
		$ret[ 1 ]	= $new_y;
		return $ret;
	}

	/**
	 * Drop in replacement for fast image resizing
	 * http://www.php.net/manual/en/function.imagecopyresampled.php#77679
	 * 
	 * Plug-and-Play fastimagecopyresampled function replaces much 
	 * slower imagecopyresampled.  Just include this function and change all 
	 * "imagecopyresampled" references to "fastimagecopyresampled".
	 * Typically from 30 to 60 times faster when reducing high resolution
	 * images down to thumbnail size using the default quality setting.
	 * @author	Tim Eckel
	 * @version	1.1
	 * @date		09/07/07
	 * Project: FreeRingers.net - Freely distributable - 
	 * These comments must remain.
	 * Optional "quality" parameter (defaults is 3). Fractional values 
	 * are allowed, for example 1.5. Must be greater than zero.
	 * Between 0 and 1 = Fast, but mosaic results, closer to 0 increases
	 * the mosaic effect.
	 * 1 = Up to 350 times faster. Poor results, looks very similar to
	 * imagecopyresized.
	 * 2 = Up to 95 times faster.  Images appear a little sharp, some prefer
	 * this over a quality of 3.
	 * 3 = Up to 60 times faster.  Will give high quality smooth results 
	 * very close to imagecopyresampled, just faster.
	 * 4 = Up to 25 times faster.  Almost identical to imagecopyresampled 
	 * for most images.
	 * 5 = No speedup. Just uses imagecopyresampled, no advantage over
	 * imagecopyresampled.
	 */
	public function fastimagecopyresampled (&$dst_image, $src_image, $dst_x, $dst_y, $src_x, $src_y, $dst_w, $dst_h, $src_w, $src_h, $quality = 3) {
		if (empty($src_image) || empty($dst_image) || $quality <= 0) { return false; }
		if ($quality < 5 && (($dst_w * $quality) < $src_w || ($dst_h * $quality) < $src_h)) {
		$temp = imagecreatetruecolor ($dst_w * $quality + 1, $dst_h * $quality + 1);
		imagecopyresized ($temp, $src_image, 0, 0, $src_x, $src_y, $dst_w * $quality + 1, $dst_h * $quality + 1, $src_w, $src_h);
		imagecopyresampled ($dst_image, $temp, $dst_x, $dst_y, 0, 0, $dst_w, $dst_h, $dst_w * $quality, $dst_h * $quality);
		imagedestroy ($temp);
		} else imagecopyresampled ($dst_image, $src_image, $dst_x, $dst_y, $src_x, $src_y, $dst_w, $dst_h, $src_w, $src_h);
		return true;
	}

	/**
	 * Exit with an error string in a nicely form HTML page
	 * @param	string	$message	The text to display
	 * @access	public
	 */
	function exit_error ( $message ) {
		print $this->_html_head();
		print   "<div class=\"error\">"
					. "<h2>Error!</h2>\n"
					. "<p>$message</p>\n"
				. "</div>\n";
		print $this->_html_foot();
		exit();	
	}

	/**
	 * Exit with an OK message
	 * @param	string	$message	The text to display
	 * @access	public
	 */
	function exit_ok ( $message ) {
		print $this->_html_head();
		print   "<div class=\"good\">"
					. "<h2>Success!</h2>\n"
					. "<p>$message</p>\n"
				. "</div>\n";
		print $this->_html_foot();
		exit();	
	}

	/**
	 * Compare two objects by their 'date' property
	 * @param	mixed	 	$a		Object A
	 * @param	mixed	 	$b		Object B
	 * @return 	integer			strcmp() result
	 * @access	private
	 */
	function _cmp_date ( $a, $b ) {
		return strnatcmp( $a["date"], $b["date"] );
	}
	
	/**
	 * Compare two objects by their 'mtime' property
	 * @param	mixed	 	$a			Object A
	 * @param	mixed	 	$b			Object B
	 * @return 	integer				strcmp() result
	 * @access	private
	 */
	function _cmp_mtime ( $a, $b ) {
		return strcmp( $a["mtime"], $b["mtime"] );
	}
	
	/**
	 * Compare two objects by their 'views' property
	 * @param	mixed	 	$a			Object A
	 * @param	mixed	 	$b			Object B
	 * @return 	integer				strcmp() result
	 * @access	private
	 */
	function _cmp_views ( $a, $b ) {
		return strnatcmp( $a["views"], $b["views"] );
	}

	/**
	 * Build a radio button for an HTML form
	 * @param 	string 	$name		The name of the element
	 * @param 	string 	$value	The value of the element
	 * @param 	string 	$label	Optional label string to wrap around
	 * @return 	string 				Built HTML element
	 * @access	public
	 */
	public function html_radio ( $name, $value, $label = false ) {
		
		$html	= "<input "
				. "type=\"radio\" "
				. "name=\"$name\" "
				. "value=\"$value\" "
				. ( ( $this->_post[ $name ] == $value ) ? "checked " : "" )
				. ">";
		
		// Optional label	
		if ( $label ) { $html .= "<label for=\"$name\">$label</label>"; }
		
		return $html;		
	}

	/**
	 * Build a checkbox for an HTML form
	 * @param 	string 	$name		The name of the element
	 * @param 	string 	$value	The value of the element
	 * @param 	string 	$label	Optional label string to wrap around
	 * @return 	string 				Built HTML element
	 * @access	public
	 */
	public function html_checkbox ( $name, $value, $label = false ) {
		
		$html	= "<input "
				. "type=\"checkbox\" "
				. "name=\"$name\" "
				. "value=\"$value\" "
				. ( ( $this->_post[ $name ] == $value ) ? "checked " : "" )
				. ">";
		
		// Optional label	
		if ( $label ) { $html .= "<label for=\"$name\">$label</label>"; }
		
		return $html;		
	}
	
	/**
	 * Build a hidden element for an HTML form
	 * @param 	string 	$name		The name of the element
	 * @param 	string 	$value	The value of the element
	 * @return 	string 				Built HTML element
	 * @access	public
	 */
	public function html_hidden ( $name, $value ) {
		$html	= "<input "
				. "type=\"hidden\" "
				. "name=\"$name\" "
				. "value=\"$value\" "
				. ">";
		return $html;		
	}
	
	/**
	 * Build a submit element for an HTML form
	 * @param 	string 	$name		The name of the element
	 * @param 	string 	$value	The value of the element
	 * @return 	string 				Built HTML element
	 * @access	public
	 */
	public function html_submit ( $name, $value ) {
		$html	= "<input "
				. "type=\"submit\" "
				. "name=\"$name\" "
				. "value=\"$value\" "
				. ">";
		return $html;		
	}
	
	/**
	 * Move a number of items in the source list by the given direction
	 * @param	array 	$list		Source list of items
	 * @param	array 	$search	List of items to move
	 * @param	integer	$dir		<0 to move up, >0 to move down
	 * @return 	array 				The re-ordered list
	 * @access	public
	 */
	public function list_arrange ( $list, $search, $dir = -1 ) {
		
		// If they're moving down we reverse the list order
		if ( $dir > 0 ) { $search = array_reverse( $search ); }
		
		$list_size	= count( $list );
		
		// Scan through the list for each item
		foreach ( $search as $item ) { 
			
			// Find it in the list
			$places	= array_keys( $list, $item );
			$place	= $places[ 0 ];
			
			$swap		= -1;
			
			// If we're moving up we can't move up off the top
			if ( $dir < 0 ) {
				$swap	= max( 0, ( $place - 1 ) );
			}
			
			// If we're moving down we can't go off the end
			else if ( $dir > 0 ) {
				$swap	= min( ( $list_size - 1 ), ( $place + 1 ) );
			}
			
			// If we have a valid item to swap with
			if ( $swap >= 0 ) {
				$hold					= $list[ $place ];
				$list[ $place ]	= $list[ $swap ];
				$list[ $swap ]		= $hold;
			}
		}

		// All done!
		return $list;
	}
	
	/**
	 * Parse a string to provide a safe name for use in a form, by removing
	 * everything bar alphanumeric chars
	 * @param	string	$string	The string to parse
	 * @return 	string			 	The string, with bad chars removed
	 * @access	public
	 */
	public function parse_safe_name ( $string ) { 
		preg_match_all( "/[a-zA-Z0-9]/", $string, $match );
		return join( "", $match[0] );
	}

	/**
	 * Convert a PHP object into a JSON encoded string
	 * Borrowed from: http://www.post-hipster.com/?p=35
	 * @param	object	$obj		Object to encode
	 * @return 	string				JSON representation of the object
	 * @access	public
	 */
	public function json_encode ( $obj ) {
		
		if ( is_array( $obj ) ) {
			
			if ( $this->json_is_associative( $obj ) ) {
				
				$arr_out = array();
				foreach ( $obj as $key => $val ) {
					$arr_out[] = '"' . $key . '":' . $this->json_encode( $val );
				}
				
				return '{' . implode( ',', $arr_out ) . '}';
			} 
			else {
				$arr_out	= array();
				$ct		= count($obj);
				for ( $j = 0; $j < $ct; $j++ ) {
					$arr_out[] = $this->json_encode( $obj[ $j ] );
				}
				return '[' . implode( ',', $arr_out ) . ']';
			}
		} 
		else {
			if ( is_int( $obj ) ) {
				return $obj;
			} 
			else {
				$str_out = stripslashes( trim( $obj ) );
				$str_out = str_replace( array('"', '', '/'), 
												array('\"', '\\', '/'), 
												$str_out );
				return '"' . $str_out . '"';
			}
		}
	}
	
	/**
	 * Test if the given array is associative or not
	 * @param	array 	$array 	The array to test
	 * @return 	boolean				True or false accordingly
	 * @access	public
	 */
	public function json_is_associative($array) {
		$count = count( $array );
		for ( $i = 0 ; $i < $count; $i++ ) {
			if ( !( array_key_exists( $i, $array ) ) ) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Add a message to the debug log
	 * @param 	string	$msg		Message to add
	 * @return 	boolean			 	True, always
	 * @access	public
	 */
	public function debug ( $msg ) {
		array_push( $this->_debug_log, trim( $msg ) );
		return true;
	}

	/**
	 * The HTML header we use
	 * @return 	string				Chunk of HTML
	 * @access	private
	 */
	private function _html_head ( ) {
		
		$slide_hdr		= "";
		$css_hdr			= "";
		$custom_hdr		= "";
		$rebuild_hdr	= "";
		
		// If we have a home page link, we should add it
		$home_url		= $this->params["home_url"];
		$home_link		= $this->params["home_link"];
		$self				= $this->params["self"];

		// The top album listing page
		$link		= "<a href=\"$self\" accesskey=\"h\">"
					. $this->params["page_title"]
					. "</a>";
		array_unshift( $this->_html_title, $link );
		
		// Do we have an optional home page URL?
		if ( ( $home_url != "" ) && ( $home_link != "" ) ) {
			$link	= "<a href=\"$home_url\" title=\"$home_link\">"
					. $home_link
					. "</a>";
			array_unshift( $this->_html_title, $link );
		}
		
		// Are we in slideshow mode?
		if ( $this->_in_show ) {
			$slide_time		= $this->params["slide_time"];
			$slide_place	= $this->_get["place"] + 1;
			$slide_url		= $this->params["album_url"]
								. "&place=" . $slide_place
								. "&slideshow=" . $this->_get["slideshow"]
								. "";
			$slide_hdr		= "\t<meta "
								. "http-equiv=\"refresh\" "
								. "content=\"$slide_time;"
								. "url=$slide_url\">\n";
		}
		
		// Create the page title
		$page_title = join( $this->params["title_join"], $this->_html_title );
		$page_nav 	= join( $this->params["nav_join"], $this->_html_nav );
		$page_admin	= join( $this->params["nav_join"], $this->_html_admin );
		
		// Do we have some custom CSS?
		if ( $this->params["custom_css"] ) { 
			$css_hdr	= "\t<link rel=\"stylesheet\" "
						. "href=\"" . $this->params["custom_css"] . "\" "
						. "title=\"Custom CSS\">\n";
		}
		
		// What about a customised header?
		if ( $this->params["custom_header"] ) {
			$custom_hdr	= join( "", file( $this->params["custom_header"] ) );
		}
		
		// If there are albums to rebuild?
		if ( count( $this->params["album_rebuild"] ) > 0 ) {
			
			$rebuild_str	= "<div class=\"message\">"
								. "<h2>Previews and Thumbnails</h2>\n"
								. "<p>"
									. "The preview and thumbnail caches need "
									. "rebuilding for the following albums: "
									. "(click on the album name to start "
									. "the rebuild)"
								. "</p>\n"
								. "<ul>\n";
								
			foreach ( $this->params["album_rebuild"] as $album => $val ) {
				$the_url			= $self
									. "?album=" . urlencode( stripslashes( $album ) )
									. "&rebuild=cache"
									. "";
				$rebuild_str	.= "<li><a href=\"$the_url\">$album</a></li>";
			}
			$rebuild_str	.= "<ul>\n"
								. "</div>\n"
								. "";
		}

		$str	= '<html>
<head>
	<title>' . strip_tags( $page_title ) . '</title>
	<meta http-equiv="Content-Type" content="text/html" />
	<meta name="keywords" content="microalbum,image,gallery,php,gaby,vanhegan" />
	<meta name="description" content="A fast, lightweight PHP image gallery" />
	<meta name="author" content="Gaby Vanhegan" />
' . $slide_hdr . '</head>
<style type="text/css">' . $this->_html_css() . '</style>
' . $css_hdr . '<body>
' . $custom_hdr . '
<div id="microalbum" class="all">
<div class="header">
	<h1>' . $page_title . '</h1>
</div>
<div class="content">
' . $rebuild_str . '
	<p class="nav">' . $page_nav . '</p>
	<p class="admin">' . $page_admin . '</p>
' . $message . '

	<div class="images">';
	
	return $str;
	}
	
	/**
	 * The HTML footer we use
	 * @return 	string				Chunk of HTML
	 * @access	private
	 */
	private function _html_foot ( ) {
		
		$custom_foot	= "";
		if ( $this->_params["custom_footer"] ) {
			$custom_foot	= join( "", file( $this->params["custom_footer"] ) );
		}

		$log		= "";

		if ( $this->params["debug"] > 0 ) {
			
			if ( $this->params["debug"] == 1 ) {
				$log	= "<!--\n"
						. "Debug Log\n"
						. join( "\n", $this->_debug_log ) 
						. "-->\n";
			}
			else if ( $this->params["debug"] == 2 ) {
				$log	= "<div class=\"debug\">"
						. "<h2>Debug Log</h2>\n"
						. "<ol>\n"
						. "<li>"
							. join( "</li>\n<li>", $this->_debug_log ) 
						. "</li>\n"
						. "</ol>\n"
						. "</div>\n";
			}
			else if ( $this->params["debug"] == 3 ) {
				$log	= "<div class=\"debug\">"
						. "<pre>"
						. "Debug Log\n"
						. join( "\n", $this->_debug_log ) 
						. "</pre>\n"
						. "</div>\n";
			}
		}

		// Calculate how long the page took to build
		list( $usec, $sec )	= explode( " ", microtime() );
		$time_now				= ( (float)$usec + (float)$sec );
		$time_then				= $this->_page_time;
		$page_build_time		= sprintf( "%.2f", $time_now - $time_then );

		$str	= "</div>\n"
				. "</div>\n"
  				. "<div class=\"footer\">"
					. "<p>"
						. "<a href=\"http://vanhegan.net/microalbum/\">"
							. "MicroAlbum " . $this->params["version"]
						. "</a>"
					. "</p>\n"
				. "</div>\n"
				. "</div>\n"
				. "<!-- Page Build: $page_build_time secs -->\n"
				. $custom_foot
				. $log
				. "</body>\n"
				. "</html>\n"
				. "";
		
		return $str;
	}
	
	/**
	 * Get the CSS associated with the default template
	 * @return 	string			Chunk of CSS
	 * @access	private
	 */
	private function _html_css ( ) { 
		$str	= '
div#microalbum { margin: 0px; padding: 0px; background: #ddd; border: 1px solid #333; }
div#microalbum, input { font-family: Arial, Helvetica, sans-serif; font-size: 10pt; }
.header, .footer, .content, .error, .message { padding:6px 16px 6px 16px; }
div.header { border-bottom: 1px solid black; }
div.footer { border-top: 1px solid black; }
div.content { background: white; }
div.error { border: 3px solid red; padding: 6px; }
div.message { border: 3px solid #262; padding: 6px; }
div.debug { border: 1px solid #bbb; padding: 6px; margin-top: 6px; }
p, h1, h2, h3 { margin: 2px; padding: 2px; }
h1 { font-size: 14pt; font-weight: normal; }
h2 { font-size: 14pt; font-weight: normal; }
h3 { font-size: 14pt; font-weight: normal; }
td { text-align: center; vertical-align: bottom; }
p, li, ul, td { font-size: 10pt; }
img { border: 1px solid black; }
a img { text-decoration: none; }
a { border: 0px; color: #262; }
a:hover { color: #c33; }
#microalbum .normal td { text-align: left; vertical-align: top; }
#microalbum div.thumb { padding: 8px 2px 8px 2px; }
#microalbum div.preview { text-align: center; padding: 16px 4px 16px 4px; }
#microalbum p.caption { font-size: 14pt; padding-top: 6px; }
#microalbum .subcaption { font-size: 10pt; }
#microalbum .good { color: #66d; }
#microalbum .admin { float: right; clear: none; }
#microalbum .nav { float: left; clear: none; }
#microalbum .pagestat { float: right; }
#microalbum .images { clear: both; }';
		return $str;
	}

}

?>